__version__ = "0.1.0"
__author__ = 'Niklas Wulff'
__contributors__ = 'Fabia Miorelli, Benjamin Fuchs'
__email__ = 'niklas.wulff@dlr.de'
__credits__ = 'German Aerospace Center (DLR)'
__license__ = 'BSD-3-Clause'

import os
import yaml
import shutil
import pathlib
import click
import vencopy


# @click.option("--dir", default='', help='Specify separate aboslute path where the user folder should be set up')
@click.command()
@click.option("--name", default='vencopy_user', prompt="Please type the user folder name:",
              help="The folder name of the vencopy user folder created at command line working directory")
@click.option("--tutorials", default='true', help='Specify if tutorials should be copied to the user folder on set up. '
                                                  'Defaults to true')
def create(name: str, tutorials: bool):
    """VencoPy folder set up after installation"""
    cwd = pathlib.Path(os.getcwd())
    target = cwd / name
    source = pathlib.Path(vencopy.__file__).parent.resolve()
    if not os.path.exists(target):
        os.mkdir(target)
        setupFolders(src=source, trg=target, tutorials=tutorials)
        click.echo(f'VencoPy user folder created under {target}')
    elif os.path.exists(target) and not os.path.exists(target / 'run.py'):
        setupFolders(src=source, trg=target, tutorials=tutorials)
        click.echo(f'VencoPy user folder filled under {target}')
    else:
        click.echo('File run.py already exists in specified folder, for a new setup please specify a non-existent '
                   'folder name')


def setupFolders(src: pathlib.Path, trg: pathlib.Path, tutorials: bool):
    """
    Setup function to create a vencopy user folder and to copy run, config and tutorial files from the package source

    :param src: Absolute path to the vencopy package source folder
    :param trg: Absolute path to the vencopy user folder
    :param tutorials: Boolean, if true (default) tutorials are being copied from package source to user folder
    :return: None
    """
    os.mkdir(trg / 'output')
    os.mkdir(trg / 'output' / 'dataParser')
    os.mkdir(trg / 'output' / 'tripDiaryBuilder')
    os.mkdir(trg / 'output' / 'gridModeler')
    os.mkdir(trg / 'output' / 'flexEstimator')
    os.mkdir(trg / 'output' / 'evaluator')
    shutil.copy(src=src / 'run.py', dst=trg)
    shutil.copytree(src=src / 'config', dst=trg / 'config')
    if tutorials:
        shutil.copytree(src=src / 'tutorials', dst=trg / 'tutorials')
    updateLocalPathCfg(newVPRoot=trg)


def updateLocalPathCfg(newVPRoot: pathlib.Path):
    with open(newVPRoot / 'config' / 'localPathConfig.yaml') as f:
        pathCfg = yaml.load(f, Loader=yaml.SafeLoader)

    pathCfg['pathAbsolute']['vencoPyRoot'] = newVPRoot.__str__()

    with open(newVPRoot / 'config' / 'localPathConfig.yaml', 'w') as f:
        yaml.dump(pathCfg, f)


if __name__ == '__main__':
    create()






__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff'
__contributors__ = 'Fabia Miorelli, Parth Butte'
__email__ = 'Niklas.Wulff@dlr.de'
__birthdate__ = '03.11.2019'
__status__ = 'prod'  # options are: dev, test, prod
__license__ = 'BSD-3-Clause'


#----- imports & packages ------
from pathlib import Path
import functools
import warnings
import os
import yaml
import pandas as pd
import numpy as np
from random import seed, random

from vencopy.scripts.globalFunctions import createFileString, mergeVariables, calculateWeightedAverage, \
    writeProfilesToCSV


class FlexEstimator:
    def __init__(self, globalConfig: dict, flexConfig : dict, evaluatorConfig: dict, ParseData,
                 datasetID: str='MiD17'):
        """
        Class to estimate uncontrolled charging, electricity drain, grid connection, auxilliary fuel, SOC min and
        SOC max profiles based on hourly driving and boolean grid connection profiles. Requires a .xlsx file specifying
        a global value for specific electric consumption (in kWh/ 100 km) and a global rated capacity of considered
        charging stations. Automatically compiles input file names from the filekeys "inputDataDriveProfiles" /
        "inputDataPlugProfiles", runlabel (as specified in globalConfig) and datasetID (as given on instantiation). The
        number of iterations for SOC max and min calculations is defined in the function self.baseProfielCalculation as
        parameter of calcChargeMinProfiles() and calcChargeMaxProfiles(). A complete estimation consists of the
        following calls also specified in self.run():
        self.baseProfileCalculation()
        self.filter()
        self.aggregate()
        self.correct()
        self.normalize()
        self.writeOut()

        :param globalConfig: Dictionary used for relative paths, filenames and runlabels
        :param flexConfig: So far not used
        :param evaluatorConfig: DEPRECATED, may be cleared soon
        :param ParseData: Class instance of type DataParser
        :param datasetID: String used for file name composition on input files
        """

        self.globalConfig = globalConfig
        self.flexConfig = flexConfig
        self.evaluatorConfig = evaluatorConfig
        self.hourVec = range(self.globalConfig['numberOfHours'])
        self.datasetID = datasetID
        self.driveProfilesIn, self.plugProfilesIn = self.readVencoInput(datasetID=datasetID)
        self.mergeDataToWeightsAndDays(ParseData)
        self.weights = self.indexWeights(weights=self.driveProfilesIn.loc[:, ['hhPersonID', 'tripStartWeekday',
                                                                              'tripWeight']])
        self.driveProfiles, self.plugProfiles = self.indexDriveAndPlugData(driveData=self.driveProfilesIn,
                                                                           plugData=self.plugProfilesIn,
                                                                           dropIdxLevel='tripWeight',
                                                                           nHours=self.globalConfig['numberOfHours'])
        self.scalarsProc = self.procScalars(driveProfiles_raw=self.driveProfilesIn,
                                            plugProfiles_raw=self.plugProfilesIn, driveProfiles=self.driveProfiles,
                                            plugProfiles=self.plugProfiles)

        #  Future release: In a future release, a composition approach with a dataclass based
        #  encapsulation will be pursued here. This will also make it easy to communicate the data to the outside of
        #  the class as the fields are then no longer directly in the main processing class.

        # Base profile attributes
        self.drainProfiles = None
        self.chargeProfiles = None
        self.chargeMaxProfiles = None
        self.chargeProfilesUncontrolled = None
        self.auxFuelDemandProfiles = None
        self.chargeMinProfiles = None

        # Filtering attributes
        self.randNoPerProfile = None
        self.profileSelectors = None
        self.electricPowerProfiles = None
        self.plugProfilesCons = None
        self.electricPowerProfilesCons = None
        self.chargeProfilesUncontrolledCons = None
        self.auxFuelDemandProfilesCons = None
        self.profilesSOCMinCons = None
        self.profilesSOCMaxCons = None

        # Aggregation attributes
        self.plugProfilesAgg = None
        self.electricPowerProfilesAgg = None
        self.chargeProfilesUncontrolledAgg = None
        self.auxFuelDemandProfilesAgg = None
        self.plugProfilesWAgg = None
        self.electricPowerProfilesWAgg = None
        self.chargeProfilesUncontrolledWAgg = None
        self.auxFuelDemandProfilesWAgg = None
        self.plugProfilesWAggVar = None
        self.electricPowerProfilesWAggVar = None
        self.chargeProfilesUncontrolledWAggVar = None
        self.auxFuelDemandProfilesWAggVar = None

        self.socMin = None
        self.socMax = None
        self.socMinVar = None
        self.socMaxVar = None

        # Correction attributes
        self.chargeProfilesUncontrolledCorr = None
        self.electricPowerProfilesCorr = None
        self.auxFuelDemandProfilesCorr = None

        # Normalization attributes
        self.socMinNorm = None
        self.socMaxNorm = None

        # Attributes for write-out and plotting
        self.profileDictOut = {}

        print('Flex Estimator initialization complete')



    def readInputCSV(self, filePath) -> pd.DataFrame:
        """
        Reads input and cuts out value columns from a given CSV file.

        :param filePath: Relative file path to CSV file
        :return: Pandas dataframe with raw input from CSV file
        """
        inputRaw = pd.read_csv(filepath_or_buffer=filePath, header=0)
        inputData = inputRaw.loc[:, ~inputRaw.columns.str.match('Unnamed')]
        inputData = inputData.convert_dtypes()
        return inputData

    def booleanMapping(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Replaces given strings with python values for true or false.

        :param df: Dataframe holding strings defining true or false values
        :return: Dataframe holding true and false
        """
        dictBol = {'WAHR': True,
                   'FALSCH': False}
        outBool = df.replace(to_replace=dictBol, value=None)
        return outBool

    def readInputBoolean(self, filePath) -> pd.DataFrame:
        """
        Wrapper function for reading boolean data from CSV.

        :param filePath: Relative path to CSV file
        :return: Returns a dataframe with boolean values
        """

        inputRaw = self.readInputCSV(filePath=filePath)
        # inputData = self.booleanMapping(inputRaw)
        return inputRaw

    def readVencoInput(self, datasetID: str) -> tuple:
        """
        Initializing action for VencoPy-specific config-file, path dictionary and data read-in. The config file has
        to be a dictionary in a .yaml file containing three categories: pathRelative, pathAbsolute and files. Each
        category must contain itself a dictionary with the pathRelative to data, functions, plots, scripts, config and
        tsConfig. Absolute paths should contain the path to the output folder. Files should contain a path to scalar input
        data, and the two timeseries files inputDataDriveProfiles and inputDataPlugProfiles.

        :param config: A yaml config file holding a dictionary with the keys 'pathRelative' and 'pathAbsolute'
        :return: Returns four dataframes: A path dictionary, scalars, drive profile data and plug profile
        data, the latter three ones in a raw data format.
        """
        print('Reading Venco input scalars, drive profiles and boolean plug profiles')

        #scalars = self.readInputScalars(self.flexConfig['inputDataScalars'])
        driveProfiles_raw = self.readInputCSV(filePath=Path(self.globalConfig['pathRelative']['diaryOutput']) /
                                              createFileString(globalConfig=self.globalConfig,
                                                               fileKey='inputDataDriveProfiles',
                                                               datasetID=datasetID))
        plugProfiles_raw = self.readInputBoolean(filePath=Path(self.globalConfig['pathRelative']['gridOutput']) /
                                                 createFileString(globalConfig=self.globalConfig,
                                                                  fileKey='inputDataPlugProfiles',
                                                                  datasetID=datasetID))

        print('There are ' + str(len(driveProfiles_raw)) + ' drive profiles and ' +
              str(len(driveProfiles_raw)) + ' plug profiles.')

        return driveProfiles_raw, plugProfiles_raw

    def procScalars(self, driveProfiles_raw, plugProfiles_raw, driveProfiles: pd.DataFrame,
                    plugProfiles: pd.DataFrame):
        """
        Calculates some scalars from the input data such as the number of hours of drive and plug profiles, the number of
        profiles etc.

        :param driveProfiles: Input drive profile input data frame with timestep specific driving distance in km
        :param plugProfiles: Input plug profile input data frame with timestep specific boolean grid connection values
        :return: Returns a dataframe of processed scalars including number of profiles and number of hours per profile
        """

        noHoursDrive = len(driveProfiles.columns)
        noHoursPlug = len(plugProfiles.columns)
        noDriveProfilesIn = len(driveProfiles)
        noPlugProfilesIn = len(plugProfiles)
        scalarsProc = {'nHoursDrive': noHoursDrive,
                       'nHoursPlug': noHoursPlug,
                       'nDriveProfilesIn': noDriveProfilesIn,
                       'nPlugProfilesIn': noPlugProfilesIn}
        if noHoursDrive == noHoursPlug:
            scalarsProc['nHours'] = noHoursDrive
        else:
            warnings.warn('Length of drive and plug input data differ! This will at the latest crash in calculating '
                          'profiles for SoC max')
        return scalarsProc

    def indexWeights(self, weights: pd.DataFrame) -> pd.DataFrame:
        """
        Reduces the dtype from string to float if possible and sets the index of the weights to align with the indices
        of drive and plug profiles (hhPersonIDs and tripStartWeekday).

        :param weights: dataframe containing the MiD trip weights of the original trips
        :return: An indexed pandas DataFrame of the MiD trip weights
        """
        weights = weights.convert_dtypes()
        return weights.set_index(['hhPersonID', 'tripStartWeekday'], drop=True)

    def findIndexCols(self, data: pd.DataFrame, nHours: int) -> list:
        """
        Identifies columns that contain index strings rather than numeric data. It does so by checking for column
        names equal to the numbers from 0 to the number of hours.

        :param data: Pandas DataFrame where index columns should be found
        :param nHours: Integer giving the analyzed number of hours
        :return: List of index columns
        """

        dataCols = [str(i) for i in range(0, nHours + 1)]
        return data.columns[~data.columns.isin(dataCols)]

    def indexProfile(self, data: pd.DataFrame, nHours: int) -> pd.DataFrame:
        """
        Takes raw data as input and indices different profiles with the specified index columns und an unstacked form.

        :param driveProfiles_raw: Dataframe of raw drive profiles in km with as many index columns as elements
            of the list in given in indices. One column represents one timestep, e.g. hour.
        :param plugProfiles_raw: Dataframe of raw plug profiles as boolean values with as many index columns
            as elements of the list in given in indices. One column represents one timestep e.g. hour.
        :param indices: List of column names given as strings.
        :return: Two indexed dataframes with index columns as given in argument indices separated from data columns
        """

        indexCols = self.findIndexCols(data=data, nHours=nHours)
        data = data.convert_dtypes()  # Reduce column data types if possible (specifically hhPersonID column to int)
        dataIndexed = data.set_index(list(indexCols))

        # Typecast column indices to int for later looping over a range
        dataIndexed.columns = dataIndexed.columns.astype(int)
        return dataIndexed

    def indexDriveAndPlugData(self, driveData: pd.DataFrame, plugData: pd.DataFrame, dropIdxLevel: str,
                              nHours: int) -> tuple:
        """
        Wrapper function for indexing drive and plug profiles so that value columns are all made up of hourly data.

        :param driveData: Hourly drive data for individual vehicles as floats
        :param plugData: Hourly plug data for individual vehicles as boolean
        :param dropIdxLevel: Column to be dropped
        :param nHours: Integer specifying the number of values columns
        :return: Tuple of indexed drive and plug profiles as pandas DataFrames
        """
        driveProfiles = self.indexProfile(data=driveData, nHours=nHours)
        plugProfiles = self.indexProfile(data=plugData, nHours=nHours)
        return driveProfiles.droplevel(dropIdxLevel), plugProfiles.droplevel(dropIdxLevel)

    def mergeDataToWeightsAndDays(self, ParseData):
        """
        Function to merging weekday and trip weight data to driving and plugging data of respective personHHIDs. It is
        assumed that trips occur on one daz and trip weights are equal for all trips of one hhPersonID

        :param ParseData: Class instance of type DataParser
        :return: None
        """

        self.driveProfilesIn = mergeVariables(data=self.driveProfilesIn, variableData=ParseData.data,
                                              variables=['tripStartWeekday', 'tripWeight'])
        self.plugProfilesIn = mergeVariables(data=self.plugProfilesIn, variableData=ParseData.data,
                                             variables=['tripStartWeekday', 'tripWeight'])

    def calcDrainProfiles(self, driveProfiles: pd.DataFrame, flexConfig: dict) -> pd.DataFrame:
        """
        Calculates electrical consumption profiles from drive profiles assuming specific consumption (in kWh/100 km)
        given in scalar input data file.

        :param driveProfiles: indexed profile file
        :param flexConfig: YAML config which holds all relative paths and filenames for flexEstimators.py
        :return: Returns a dataframe with consumption profiles in kWh/h in same format and length as driveProfiles but
        scaled with the specific consumption assumption.
        """

        return driveProfiles * flexConfig['inputDataScalars']['Electric_consumption'] / float(100)

    def calcChargeProfiles(self, plugProfiles: pd.DataFrame, flexConfig) -> pd.DataFrame:
        '''
        Calculates the maximum possible charge power based on the plug profile assuming the charge column power
        given in the scalar input data file (so far under Panschluss).

        :param plugProfiles: indexed boolean profiles for vehicle connection to grid
        :param flexConfig: YAML config which holds all relative paths and filenames for flexEstimators.py
        :return: Returns scaled plugProfile in the same format as plugProfiles.
        '''

        return plugProfiles * flexConfig['inputDataScalars']['Rated_power_of_charging_column']

    def calcChargeMaxProfiles(self, chargeProfiles: pd.DataFrame, consumptionProfiles: pd.DataFrame,
                              nIter: int) -> pd.DataFrame:
        """
        Calculates all maximum SoC profiles under the assumption that batteries are always charged as soon as they
        are plugged to the grid. Values are assured to not fall below SoC_min * battery capacity or surpass
        SoC_max * battery capacity. Relevant profiles are chargeProfile and consumptionProfile. An iteration assures
        the boundary condition of chargeMaxProfile(0) = chargeMaxProfile(len(profiles)). The number of iterations
        is given as parameter.

        :param chargeProfiles: Indexed dataframe of charge profiles.
        :param consumptionProfiles: Indexed dataframe of consumptionProfiles.
        :param flexConfig: YAML config holds all relative paths and filenames for flexEstimators.py
        :param scalarsProc: DataFrame holding information about profile length and number of hours.
        :param nIter: Number of iterations to assure that the minimum and maximum value are approximately the same
        :return: Returns an indexed DataFrame with the same length and form as chargProfiles and consumptionProfiles,
        containing single-profile SOC max values for each hour in each profile.
        """

        chargeMaxProfiles = chargeProfiles.copy()
        batCapMin = self.flexConfig['inputDataScalars']['Battery_capacity'] * \
                    self.flexConfig['inputDataScalars']['Minimum_SOC']
        batCapMax = self.flexConfig['inputDataScalars']['Battery_capacity'] * \
                    self.flexConfig['inputDataScalars']['Maximum_SOC']
        nHours = self.scalarsProc['nHours']
        for idxIt in range(nIter):
            print(f'Starting with iteration {idxIt}')
            for iHour in range(nHours):
                if iHour == 0:
                    chargeMaxProfiles[iHour] = chargeMaxProfiles[nHours - 1].where(
                        cond=chargeMaxProfiles[iHour] <= batCapMax, other=batCapMax)
                else:
                    # Calculate and append column with new SoC Max value for comparison and cleaner code
                    chargeMaxProfiles['newCharge'] = chargeMaxProfiles[iHour - 1] + \
                                                     chargeProfiles[iHour] - \
                                                     consumptionProfiles[iHour]

                    # Ensure that chargeMaxProfiles values are between batCapMin and batCapMax
                    chargeMaxProfiles[iHour] \
                        = chargeMaxProfiles['newCharge'].where(cond=chargeMaxProfiles['newCharge'] <= batCapMax,
                                                               other=batCapMax)
                    chargeMaxProfiles[iHour] \
                        = chargeMaxProfiles[iHour].where(cond=chargeMaxProfiles[iHour] >= batCapMin, other=batCapMin)

            devCrit = chargeMaxProfiles[nHours - 1].sum() - chargeMaxProfiles[0].sum()
            print(devCrit)
        chargeMaxProfiles.drop(labels='newCharge', axis='columns', inplace=True)
        return chargeMaxProfiles

    def calcChargeProfilesUncontrolled(self, chargeMaxProfiles: pd.DataFrame,
                                       scalarsProc: pd.DataFrame) -> pd.DataFrame:
        """
        Calculates uncontrolled electric charging based on SoC Max profiles for each hour for each profile.

        :param chargeMaxProfiles: Dataframe holding timestep dependent SOC max values for each profile.
        :param scalarsProc: VencoPy Dataframe holding meta-information about read-in profiles.
        :return: Returns profiles for uncontrolled charging under the assumption that charging occurs as soon as a
        vehicle is connected to the grid up to the point that the maximum battery SOC is reached or the connection
        is interrupted. DataFrame has the same format as chargeMaxProfiles.
        """

        chargeMaxProfiles = chargeMaxProfiles.copy()
        chargeProfilesUncontrolled = chargeMaxProfiles.copy()
        nHours = scalarsProc['nHours']

        for iHour in range(nHours):
            if iHour != 0:
                chargeProfilesUncontrolled[iHour] = (chargeMaxProfiles[iHour] - chargeMaxProfiles[iHour - 1]).where(
                    cond=chargeMaxProfiles[iHour] >= chargeMaxProfiles[iHour - 1], other=0)

        # set value of uncontrolled charging for first hour to average between hour 1 and hour 23
        # because in calcChargeMax iteration the difference is minimized.
        chargeProfilesUncontrolled[0] = \
            (chargeProfilesUncontrolled[1] + chargeProfilesUncontrolled[nHours - 1]) / 2
        return chargeProfilesUncontrolled

    def calcDriveProfilesFuelAux(self, chargeMaxProfiles: pd.DataFrame, chargeProfilesUncontrolled: pd.DataFrame,
                                 driveProfiles: pd.DataFrame, flexConfig,
                                 scalarsProc: pd.DataFrame) -> pd.DataFrame:
        # FixMe: alternative vectorized format for looping over columns? numpy, pandas: broadcasting-rules
        """
         Calculates necessary fuel consumption profile of a potential auxilliary unit (e.g. a gasoline motor) based
        on gasoline consumption given in scalar input data (in l/100 km). Auxilliary fuel is needed if an hourly
        mileage is higher than the available SoC Max in that hour.

        :param chargeMaxProfiles: Dataframe holding hourly maximum SOC profiles in kWh for all profiles
        :param chargeProfilesUncontrolled: Dataframe holding hourly uncontrolled charging values in kWh/h for all profiles
        :param driveProfiles: Dataframe holding hourly electric driving demand in kWh/h for all profiles.
        :param flexConfig: YAML config which holds all relative paths and filenames for flexEstimators.py
        :param scalarsProc: Dataframe holding meta-infos about the input
        :return: Returns a DataFrame with single-profile values for back-up fuel demand in the case a profile cannot
        completely be fulfilled with electric driving under the given consumption and battery size assumptions.
        """

        # Future release:
        # the hardcoding of the column names can cause a lot of problems for people later on if we do not ship
        # the date with the tool. I would recommend to move these column names to a config file similar to i18n
        # strategies
        consumptionPower = flexConfig['inputDataScalars']['Electric_consumption']
        consumptionFuel = flexConfig['inputDataScalars']['Fuel_consumption']

        # initialize data set for filling up later on
        driveProfilesFuelAux = chargeMaxProfiles.copy()
        nHours = scalarsProc['nHours']

        for iHour in range(nHours):
            if iHour != 0:
                driveProfilesFuelAux[iHour] = (consumptionFuel / consumptionPower) * \
                                              (driveProfiles[iHour] * consumptionPower / 100 -
                                               chargeProfilesUncontrolled[iHour] -
                                               (chargeMaxProfiles[iHour - 1] - chargeMaxProfiles[iHour]))

        # Setting value of hour=0 equal to the average of hour=1 and last hour
        driveProfilesFuelAux[0] = (driveProfilesFuelAux[nHours - 1] + driveProfilesFuelAux[1]) / 2
        driveProfilesFuelAux = driveProfilesFuelAux.astype(float).round(4)
        return driveProfilesFuelAux

    def calcChargeMinProfiles(self, chargeProfiles: pd.DataFrame, consumptionProfiles: pd.DataFrame,
                              driveProfilesFuelAux: pd.DataFrame, nIter: int = 3) -> pd.DataFrame:
        """
        Calculates minimum SoC profiles assuming that the hourly mileage has to exactly be fulfilled but no battery charge
        is kept inspite of fulfilling the mobility demand. It represents the minimum charge that a vehicle battery has to
        contain in order to fulfill all trips.
        An iteration is performed in order to assure equality of the SoCs at beginning and end of the profile.

        :param chargeProfiles: Charging profiles with techno-economic assumptions on connection power.
        :param consumptionProfiles: Profiles giving consumed electricity for each trip in each hour assuming specified
            consumption.
        :param driveProfilesFuelAux: Auxilliary fuel demand for fulfilling trips if purely electric driving doesn't suffice.
        :param scalarsProc: Number of profiles and number of hours of each profile.
        :param nIter: Gives the number of iterations to fulfill the boundary condition of the SoC equalling in the first
            and in the last hour of the profile.
        :return: Returns an indexed DataFrame containing minimum SOC values for each profile in each hour in the same
            format as chargeProfiles, consumptionProfiles and other input parameters.
        """
        chargeMinProfiles = chargeProfiles.copy()
        batCapMin = self.flexConfig['inputDataScalars']['Battery_capacity'] * \
                    self.flexConfig['inputDataScalars']['Minimum_SOC']
        batCapMax = self.flexConfig['inputDataScalars']['Battery_capacity'] * \
                    self.flexConfig['inputDataScalars']['Maximum_SOC']
        consElectric = self.flexConfig['inputDataScalars']['Electric_consumption']
        consGasoline = self.flexConfig['inputDataScalars']['Fuel_consumption']
        nHours = self.scalarsProc['nHours']
        for idxIt in range(nIter):
            for iHour in range(nHours):
                if iHour == nHours - 1:
                    chargeMinProfiles[iHour] = chargeMinProfiles[0].where(cond=batCapMin <= chargeMinProfiles[iHour],
                                                                          other=batCapMin)
                else:
                    # Calculate and append column with new SOC Max value for comparison and nicer code
                    chargeMinProfiles['newCharge'] = chargeMinProfiles[iHour + 1] + \
                                                     consumptionProfiles[iHour + 1] - \
                                                     chargeProfiles[iHour + 1] - \
                                                     (driveProfilesFuelAux[iHour + 1] * consElectric / consGasoline)

                    # Ensure that chargeMinProfiles values are between batCapMin and batCapMax
                    chargeMinProfiles[iHour] \
                        = chargeMinProfiles['newCharge'].where(cond=chargeMinProfiles['newCharge'] >= batCapMin,
                                                               other=batCapMin)
                    chargeMinProfiles[iHour] \
                        = chargeMinProfiles[iHour].where(cond=chargeMinProfiles[iHour] <= batCapMax, other=batCapMax)

            devCrit = chargeMinProfiles[nHours - 1].sum() - chargeMinProfiles[0].sum()
            print(devCrit)
        chargeMinProfiles.drop(labels='newCharge', axis='columns', inplace=True)
        return chargeMinProfiles

    def baseProfileCalculation(self):
        """
        Wrapper function for first part of flexibility estimation calculating the six resulting profiles for all
        individual vehicles. The number of iterations for calcChargeMaxProfiles() and calcChargeMinProfiles() can be
        specified here.

        :return: None
        """

        self.drainProfiles = self.calcDrainProfiles(driveProfiles=self.driveProfiles, flexConfig=self.flexConfig)
        self.chargeProfiles = self.calcChargeProfiles(plugProfiles=self.plugProfiles, flexConfig=self.flexConfig)
        self.chargeMaxProfiles = self.calcChargeMaxProfiles(chargeProfiles=self.chargeProfiles,
                                                            consumptionProfiles=self.drainProfiles, nIter=3)
        # self.connectionType = self.assignConnectionType()
        self.chargeProfilesUncontrolled = self.calcChargeProfilesUncontrolled(chargeMaxProfiles=self.chargeMaxProfiles,
                                                                              scalarsProc=self.scalarsProc)
        self.auxFuelDemandProfiles = self.calcDriveProfilesFuelAux(chargeMaxProfiles=self.chargeMaxProfiles,
                                                            chargeProfilesUncontrolled=self.chargeProfilesUncontrolled,
                                                                   driveProfiles=self.driveProfiles,
                                                                   flexConfig=self.flexConfig,
                                                                   scalarsProc=self.scalarsProc)
        self.chargeMinProfiles = self.calcChargeMinProfiles(chargeProfiles=self.chargeProfiles,
                                                            consumptionProfiles=self.drainProfiles,
                                                            driveProfilesFuelAux=self.auxFuelDemandProfiles)  # potentially specify nIter here

        print(f'Base profile calculation complete for dataset {self.datasetID}')

    def createRandNo(self, driveProfiles: pd.DataFrame, setSeed=1):
        """
        Creates a random number between 0 and 1 for each profile based on driving profiles.

        :param driveProfiles: Dataframe holding hourly electricity consumption values in kWh/h for all profiles
        :param setSeed: Seed for reproducing stochasticity. Scalar number.
        :return: Returns an indexed series with the same indices as dirveProfiles with a random number between 0 and 1 for
        each index.
        """

        idxData = driveProfiles.copy()
        seed(setSeed)  # seed random number generator for reproducibility
        idxData['randNo'] = np.random.random(len(idxData))
        idxData['randNo'] = [random() for _ in
                             range(len(idxData))]  # generate one random number for each profile / index
        randNo = idxData.loc[:, 'randNo']
        return randNo

    def calcProfileSelectors(self, chargeProfiles: pd.DataFrame,
                             consumptionProfiles: pd.DataFrame,
                             driveProfiles: pd.DataFrame,
                             driveProfilesFuelAux: pd.DataFrame,
                             randNos: pd.DataFrame,
                             fuelDriveTolerance,
                             isBEV: bool) -> pd.DataFrame:
        """
        This function calculates two filters. The first filter, filterCons, excludes profiles that depend on auxiliary
        fuel with an option of a tolerance (bolFuelDriveTolerance) and those that don't reach a minimum daily average for
        mileage (bolMinDailyMileage).
        A second filter filterDSM excludes profiles where charging throughout the day supplies less energy than necessary
        for the respective trips (bolConsumption) and those where the battery doesn't suffice the mileage (bolSuffBat).

        :param chargeProfiles: Indexed DataFrame giving hourly charging profiles
        :param consumptionProfiles: Indexed DataFrame giving hourly consumption profiles
        :param driveProfiles:  Indexed DataFrame giving hourly electricity demand profiles for driving.
        :param driveProfilesFuelAux: Indexed DataFrame giving auxiliary fuel demand.
        :param randNos: Indexed Series giving a random number between 0 and 1 for each profiles.
        :param fuelDriveTolerance: Give a threshold value how many liters may be needed throughout the course of a day
        in order to still consider the profile.
        :param isBEV: Boolean value. If true, more 2030 profiles are taken into account (in general).
        :return: The bool indices are written to one DataFrame in the DataManager with the columns randNo, indexCons and
        indexDSM and the same indices as the other profiles.
        """

        boolBEV = self.flexConfig['inputDataScalars']['Is_BEV?']
        minDailyMileage = self.flexConfig['inputDataScalars']['Minimum_daily_mileage']
        batSize = self.flexConfig['inputDataScalars']['Battery_capacity']
        socMax = self.flexConfig['inputDataScalars']['Maximum_SOC']
        socMin = self.flexConfig['inputDataScalars']['Minimum_SOC']
        filterCons = driveProfiles.copy()
        filterCons['randNo'] = randNos
        filterCons['bolFuelDriveTolerance'] = driveProfilesFuelAux.sum(axis='columns') * boolBEV < fuelDriveTolerance
        filterCons['bolMinDailyMileage'] = driveProfiles.sum(axis='columns') > \
                                           (2 * randNos * minDailyMileage + (1 - randNos) * minDailyMileage * isBEV)
        filterCons['indexCons'] = filterCons.loc[:, 'bolFuelDriveTolerance'] & filterCons.loc[:, 'bolMinDailyMileage']
        filterCons['bolConsumption'] = consumptionProfiles.sum(axis=1) < chargeProfiles.sum(axis=1)
        filterCons['bolSuffBat'] = consumptionProfiles.sum(axis=1) < batSize * (socMax - socMin)
        filterCons['indexDSM'] = filterCons['indexCons'] & filterCons['bolConsumption'] & filterCons['bolSuffBat']

        print('There are ' + str(sum(filterCons['indexCons'])) + ' considered profiles and ' + \
              str(sum(filterCons['indexDSM'])) + ' DSM eligible profiles.')
        filterCons_out = filterCons.loc[:, ['randNo', 'indexCons', 'indexDSM']]
        return filterCons_out

    def calcElectricPowerProfiles(self, consumptionProfiles: pd.DataFrame, driveProfilesFuelAux: pd.DataFrame,
                                  filterCons: pd.DataFrame, filterIndex) -> pd.DataFrame:
        """
        Calculates electric power profiles that serve as outflow of the fleet batteries.

        :param consumptionProfiles: Indexed DataFrame containing electric vehicle consumption profiles.
        :param driveProfilesFuelAux: Indexed DataFrame containing
        :param flexConfig: YAML config which holds all relative paths and filenames for flexEstimators.py
        :param filterCons: Dataframe containing one boolean filter value for each profile
        :param scalarsProc: Dataframe containing meta information of input profiles
        :param filterIndex: Can be either 'indexCons' or 'indexDSM' so far. 'indexDSM' applies stronger filters and results
        are thus less representative.
        :return: Returns electric demand from driving filtered and aggregated to one fleet.
        """

        consumptionPower = self.flexConfig['inputDataScalars']['Electric_consumption']
        consumptionFuel = self.flexConfig['inputDataScalars']['Fuel_consumption']
        indexCons = filterCons.loc[:, 'indexCons']
        indexDSM = filterCons.loc[:, 'indexDSM']
        nHours = self.scalarsProc['nHours']
        electricPowerProfiles = consumptionProfiles.copy()
        for iHour in range(nHours):
            electricPowerProfiles[iHour] = (consumptionProfiles[iHour] - driveProfilesFuelAux[iHour] *
                                            (consumptionPower / consumptionFuel))
            if filterIndex == 'indexCons':
                electricPowerProfiles[iHour] = electricPowerProfiles[iHour] * indexCons
            elif filterIndex == 'indexDSM':
                electricPowerProfiles[iHour] = electricPowerProfiles[iHour] * indexDSM.astype(int)
        return electricPowerProfiles

    def setUnconsideredBatProfiles(self, chargeMaxProfiles: pd.DataFrame, chargeMinProfiles: pd.DataFrame,
                                   filterCons: pd.DataFrame, minValue, maxValue):
        """
        Sets all profile values with filterCons = False to extreme values. For SoC max profiles, this means a value
        that is way higher than SoC max capacity. For SoC min this means usually 0. This setting is important for the
        next step of filtering out extreme values.

        :param chargeMaxProfiles: Dataframe containing hourly maximum SOC profiles for all profiles
        :param chargeMinProfiles: Dataframe containing hourly minimum SOC profiles for all profiles
        :param filterCons: Dataframe containing one boolean value for each profile
        :param minValue: Value that non-reasonable values of SoC min profiles should be set to.
        :param maxValue: Value that non-reasonable values of SoC max profiles should be set to.
        :return: Writes the two profiles files 'chargeMaxProfilesDSM' and 'chargeMinProfilesDSM' to the DataManager.
        """

        chargeMinProfilesDSM = chargeMinProfiles.copy()
        chargeMaxProfilesDSM = chargeMaxProfiles.copy()
        try:
            chargeMinProfilesDSM.loc[~filterCons['indexDSM'].astype('bool'), :] = minValue
            chargeMaxProfilesDSM.loc[~filterCons['indexDSM'].astype('bool'), :] = maxValue
        except Exception as E:
            print("Declaration doesn't work. "
                  "Maybe the length of filterCons differs from the length of chargeMaxProfiles")
            raise E
        return chargeMaxProfilesDSM, chargeMinProfilesDSM

    def filterConsProfiles(self, profile: pd.DataFrame, filterCons: pd.DataFrame, critCol) -> pd.DataFrame:
        """
        Filter out all profiles from given profile types whose boolean indices (so far DSM or cons) are FALSE.

        :param profile: Dataframe of hourly values for all filtered profiles
        :param filterCons: Identifiers given as list of string to store filtered profiles back into the DataManager
        :param critCol: Criterium column for filtering
        :return: Stores filtered profiles in the DataManager under keys given in dmgrNames
        """

        outputProfile = profile.loc[filterCons[critCol], :]
        return outputProfile

    def socProfileSelection(self, profilesMin: pd.DataFrame, profilesMax: pd.DataFrame, filter, alpha) -> tuple:
        """
        Selects the nth highest value for each hour for min (max profiles based on the percentage given in parameter
        'alpha'. If alpha = 10, the 10%-biggest (10%-smallest) value is selected, all other values are disregarded.
        Currently, in the Venco reproduction phase, the hourly values are selected independently of each other. min and max
        profiles have to have the same number of columns.

        :param profilesMin: Profiles giving minimum hypothetic SOC values to supply the driving demand at each hour
        :param profilesMax: Profiles giving maximum hypothetic SOC values if vehicle is charged as soon as possible
        :param filter: Filter method. Currently implemented: 'singleValue'
        :param alpha: Percentage, giving the amount of profiles whose mobility demand can not be fulfilled after selection.
        :return: Returns the two profiles 'socMax' and 'socMin' in the same time resolution as input profiles.
        """

        profilesMin = profilesMin.convert_dtypes()
        profilesMax = profilesMax.convert_dtypes()
        noProfiles = len(profilesMin)
        noProfilesFilter = int(alpha / 100 * noProfiles)
        if filter == 'singleValue':
            profileMinOut = profilesMin.iloc[0, :].copy()
            for col in profilesMin:
                profileMinOut[col] = min(profilesMin[col].nlargest(noProfilesFilter))

            profileMaxOut = profilesMax.iloc[0, :].copy()
            for col in profilesMax:
                profileMaxOut[col] = max(profilesMax[col].nsmallest(noProfilesFilter))

        else:
            raise ValueError('You selected a filter method that is not implemented.')
        return profileMinOut, profileMaxOut

    def normalizeProfiles(self, socMin: pd.Series, socMax: pd.Series) -> \
            tuple:
        """
        Normalizes given profiles with a given scalar reference.

        :param scalars: Dataframe containing technical assumptions e.g. battery capacity
        :param socMin: Minimum SOC profile subject to normalization
        :param socMax: Minimum SOC profile subject to normalization
        :param normReferenceParam: Reference parameter that is taken for normalization.
        This has to be given in scalar input data and is most likely the 'Battery_capacity'.
        :return: Writes the normalized profiles to the DataManager under the specified keys
        """

        normReference = self.flexConfig['inputDataScalars']['Battery_capacity']
        socMinNorm = socMin.div(float(normReference))
        socMaxNorm = socMax.div(float(normReference))
        return socMinNorm, socMaxNorm

    def filter(self):
        """
        Wrapper function to carry out filtering and selection procedures. A tolerance for needing additional fuel to
        carry out trips can be specified to keep profiles in the analyzed data basis.

        :return: None
        """

        self.randNoPerProfile = self.createRandNo(driveProfiles=self.driveProfiles)
        self.profileSelectors = self.calcProfileSelectors(chargeProfiles=self.chargeProfiles,
                                                     consumptionProfiles=self.drainProfiles,
                                                     driveProfiles=self.driveProfiles,
                                                     driveProfilesFuelAux=self.auxFuelDemandProfiles,
                                                     randNos=self.randNoPerProfile,
                                                     fuelDriveTolerance=1, isBEV=True)

        # Additional fuel consumption is subtracted from the consumption
        self.electricPowerProfiles = self.calcElectricPowerProfiles(consumptionProfiles=self.drainProfiles,
                                                                    driveProfilesFuelAux=self.auxFuelDemandProfiles,
                                                                    filterCons=self.profileSelectors, filterIndex='indexDSM')

        # Profile filtering for flow profiles
        self.plugProfilesCons = self.filterConsProfiles(profile=self.plugProfiles, filterCons=self.profileSelectors,
                                                        critCol='indexCons')
        self.electricPowerProfilesCons = self.filterConsProfiles(profile=self.electricPowerProfiles,
                                                                 filterCons=self.profileSelectors, critCol='indexCons')
        self.chargeProfilesUncontrolledCons = self.filterConsProfiles(profile=self.chargeProfilesUncontrolled,
                                                                      filterCons=self.profileSelectors,
                                                                      critCol='indexCons')
        self.auxFuelDemandProfilesCons = self.filterConsProfiles(profile=self.auxFuelDemandProfiles,
                                                                 filterCons=self.profileSelectors, critCol='indexCons')

        # Profile filtering for state profiles
        self.profilesSOCMinCons = self.filterConsProfiles(profile=self.chargeMinProfiles,
                                                          filterCons=self.profileSelectors, critCol='indexDSM')
        self.profilesSOCMaxCons = self.filterConsProfiles(profile=self.chargeMaxProfiles,
                                                          filterCons=self.profileSelectors, critCol='indexDSM')

    def aggregateProfilesMean(self, profilesIn: pd.DataFrame) -> pd.Series:
        """
        This method aggregates all single-vehicle profiles that are considered to one fleet profile.

        :param profilesIn: Dataframe of hourly values of all filtered profiles
        :return: Returns a Dataframe with hourly values for one aggregated profile
        """

        # Typecasting is necessary for aggregation of boolean profiles
        profilesIn = profilesIn.loc[~profilesIn.apply(lambda x: x.isna(), axis=0).any(axis=1), :]
        lenProfiles = len(profilesIn)
        profilesOut = profilesIn.apply(sum, axis=0) / lenProfiles
        return profilesOut

    def aggregateProfilesWeight(self, profiles: pd.DataFrame, weights: pd.DataFrame) -> pd.Series:
        """
        Aggregation of profiles considering the specific weights given for the household person IDs. No rescaling of
        the weights is carried out so far. The function calculateWeightedAverage() is specified in globalFunctions.py

        :param profilesIn: Dataframe of hourly values of all filtered profiles
        :param weights: Returns a Dataframe with hourly values considering the weight of individual trip
        :return:
        """
        profilesIn = profiles.loc[~profiles.apply(lambda x: x.isna(), axis=0).any(axis=1), :]
        weights = weights.loc[profilesIn.index, :]  # Filtering weight data to equate lengths
        return profilesIn.apply(calculateWeightedAverage, weightCol=weights['tripWeight'])

    def aggregateDiffVariable(self, data: pd.DataFrame, by: str, weights: pd.Series, hourVec: list) -> pd.Series:
        """
        A separate weighted aggregation function differentiating by the variable defined as a string in str. Weights as
        given in MiD.

        :param data: list of strings declaring the datasetIDs to be read in
        :param by: String specifyzing a variable.
        :param weights: Weight vector as given in the MiD
        :param hourVec: hour vector specifying the hours used for VencoPy analysis
        :return:
        """
        vars = set(data.index.get_level_values(by))
        ret = pd.DataFrame(index=hourVec, columns=vars)
        data = data.reset_index(level=by)
        weights = weights.loc[data.index, :].reset_index(level=by)
        for iVar in vars:
            dataSlice = data.loc[data.loc[:, by] == iVar, hourVec]
            weightSlice = weights.loc[weights.loc[:, by] == iVar, 'tripWeight']
            ret.loc[:, iVar] = dataSlice.apply(calculateWeightedAverage, weightCol=weightSlice)
        ret = ret.stack()
        ret.index.names = ['time', by]
        return ret

    def aggregate(self):
        """
        Wrapper function to aggregate profiles from individual vehicle level to fleet level. This is done in two
        categories: Aggregating to one representative weekday with 24 hours and aggregating for a representative week.
        Within these categories mean and weighted mean values are calculated for comparison. Simple means are only
        calculated for 24 hour profiles. For state of charge profiles (soc max and soc min), simple estimations of
        minimum and maximum state-of-charge profiles are carried out. This is done based on
        https://elib.dlr.de/92151/1/Dissertation_Diego_Luca_de_Tena.pdf by selecting the nth maximum or minimum value.
        See https://doi.org/10.3390/en14144349 for further explanations.
        """

        # Profile aggregation for flow profiles by averaging
        self.plugProfilesAgg = self.aggregateProfilesMean(self.plugProfilesCons)
        self.electricPowerProfilesAgg = self.aggregateProfilesMean(self.electricPowerProfilesCons)
        self.chargeProfilesUncontrolledAgg = self.aggregateProfilesMean(self.chargeProfilesUncontrolledCons)
        self.auxFuelDemandProfilesAgg = self.aggregateProfilesMean(self.auxFuelDemandProfilesCons)

        # Profile aggregation for flow profiles by averaging
        self.plugProfilesWAgg = self.aggregateProfilesWeight(profiles=self.plugProfilesCons, weights=self.weights)
        self.electricPowerProfilesWAgg = self.aggregateProfilesWeight(profiles=self.electricPowerProfilesCons,
                                                                      weights=self.weights)
        self.chargeProfilesUncontrolledWAgg = self.aggregateProfilesWeight(profiles=self.chargeProfilesUncontrolledCons,
                                                                           weights=self.weights)
        self.auxFuelDemandProfilesWAgg = self.aggregateProfilesWeight(profiles=self.auxFuelDemandProfilesCons,
                                                                      weights=self.weights)

        # Define a partial method for variable specific weight-considering aggregation to make following lines shorter
        aggDiffWeekday = functools.partial(self.aggregateDiffVariable, by='tripStartWeekday', weights=self.weights,
                                           hourVec=self.hourVec)

        # Profile aggregation for flow profiles by averaging
        self.plugProfilesWAggVar = aggDiffWeekday(data=self.plugProfilesCons)
        self.electricPowerProfilesWAggVar = aggDiffWeekday(data=self.electricPowerProfilesCons)
        self.chargeProfilesUncontrolledWAggVar = aggDiffWeekday(data=self.chargeProfilesUncontrolledCons)
        self.auxFuelDemandProfilesWAggVar = aggDiffWeekday(data=self.auxFuelDemandProfilesCons)

        # Profile aggregation for state profiles by selecting one profiles value for each hour
        self.socMin, self.socMax = self.socProfileSelection(profilesMin=self.profilesSOCMinCons,
                                                            profilesMax=self.profilesSOCMaxCons,
                                                            filter='singleValue', alpha=10)

        self.socMinVar, self.socMaxVar = self.socSelectionVar(dataMin=self.profilesSOCMinCons,
                                                              dataMax=self.profilesSOCMaxCons,
                                                              by='tripStartWeekday', filter='singleValue', alpha=10)

    def socSelectionVar(self, dataMin: pd.DataFrame, dataMax: pd.DataFrame, by: str, filter: str, alpha: int) -> tuple:
        """
        SOC selection function to aggregate state profiles from individual vehicle to fleet level.

        :param dataMin: Pandas Dataframe of hourly SOC min profiles for each household person ID
        :param dataMax: Pandas Dataframe of hourly SOC max profiles for each household person ID
        :param by: index level to differentiate selections by. Given as a string.
        :param filter: Filter method given as a string. Currently only 'singleValue' is implemented
        :param alpha: Percentile value to filter out extreme minimum and maximum soc values. E.g. 10 selects the 90th
        percentile for SOC max and the 10th percentile for SOC min values in each hour. These 24 values most likely do
        not belong to the same profile.
        :return: Returns a touple of estimated fleet socMin and socMax profiles for nHour x len(set(dataMin.loc[:, by])
        values. E.g. if running for 24 hour profiles additionally differentiating weekdays, this yields 168 values per
        resulting profile.
        """
        # socSelectionPartial = functools.partial(func=self.socProfileSelection, filter=filter, alpha=alpha)
        vars = set(dataMin.index.get_level_values(by))
        retMin = pd.DataFrame(index=self.hourVec, columns=vars)
        retMax = pd.DataFrame(index=self.hourVec, columns=vars)
        dataMin = dataMin.reset_index(level=by)
        dataMax = dataMax.reset_index(level=by)
        for iVar in vars:
            dataSliceMin = dataMin.loc[dataMin.loc[:, by] == iVar, self.hourVec]
            dataSliceMax = dataMax.loc[dataMax.loc[:, by] == iVar, self.hourVec]
            retMin.loc[:, iVar], retMax.loc[:, iVar] = self.socProfileSelection(profilesMin=dataSliceMin,
                                                                                profilesMax=dataSliceMax,
                                                                                filter='singleValue',
                                                                                alpha=10)

            # retMin.loc[:, iVar], retMax.loc[:, iVar] = socSelectionPartial(profilesMin=dataSliceMin,
            #                                                                profilesMax=dataSliceMax)

        retMin = retMin.stack()
        retMax = retMax.stack()
        retMin.index.names = ['time', by]
        retMax.index.names = ['time', by]
        return retMin, retMax

    def correctProfiles(self, profile: pd.Series, profType) -> pd.Series:
        """
        This method scales given profiles by a correction factor. It was written for VencoPy scaling consumption data
        with the more realistic ARTEMIS driving cycle.

        :param flexConfig: YAML config which holds all relative paths and filenames for flexEstimators.py
        :param profile: Dataframe of profile that should be corrected
        :param profType: A list of strings specifying if the given profile type is an electric or a fuel profile.
        profType has to have the same length as profiles.
        :return:
        """

        if profType == 'electric':
            consumptionElectric = self.flexConfig['inputDataScalars']['Electric_consumption']
            consumptionElectricCorr = self.flexConfig['inputDataScalars']['Electric_consumption_corr']
            corrFactor = consumptionElectricCorr / consumptionElectric
        elif profType == 'fuel':
            consumptionFuel = self.flexConfig['inputDataScalars']['Fuel_consumption']
            consumptionFuelCorr = self.flexConfig['inputDataScalars']['Fuel_consumption_corr']
            corrFactor = consumptionFuelCorr / consumptionFuel
        else:
            raise Exception(f'Either parameter "{profType}" is not given or not assigned to either "electric" or '
                            f'"fuel".')
        return corrFactor * profile

    def correct(self):
        """
        Wrapper function to correct all electric and fuel demand profiles with more realistic specific consumption
        values.

        :return: None
        """

        self.chargeProfilesUncontrolledCorr = self.correctProfiles(profile=self.chargeProfilesUncontrolledAgg,
                                                                   profType='electric')
        self.electricPowerProfilesCorr = self.correctProfiles(profile=self.electricPowerProfilesAgg,
                                                              profType='electric')
        self.auxFuelDemandProfilesCorr = self.correctProfiles(profile=self.auxFuelDemandProfilesAgg, profType='fuel')

    def normalize(self):
        """
        Normalization of soc profiles with regard to the battery capacity.
        """
        self.socMinNorm, self.socMaxNorm = self.normalizeProfiles(self.socMin, self.socMax)

    def writeOut(self):
        """
        Generic write-out function for estimated flexibility profiles. Profiles are all written to output/data as
        specified in the globalConfig and ammended by "vencoPyOutput", the runlabel as specified in the globalConfig
        as well as the datasetID. Output profiles are all written to one single file.

        :return: None
        """
        self.profileDictOut = dict(uncontrolledCharging=self.chargeProfilesUncontrolledWAggVar,
                                   electricityDemandDriving=self.electricPowerProfilesWAggVar,
                                   socMax=self.socMaxVar, socMin=self.socMinVar,
                                   gridConnectionShare=self.plugProfilesWAggVar,
                                   auxFuelDriveProfile=self.auxFuelDemandProfilesWAggVar)

        writeProfilesToCSV(profileDictOut=self.profileDictOut, globalConfig=self.globalConfig, singleFile=True,
                           datasetID=self.datasetID)

    def run(self):
        """
        Wrapper function for the whole flexibility estimation workflow containing the six above described wrapper
        functions.

        :return: None
        """
        self.baseProfileCalculation()
        self.filter()
        self.aggregate()
        self.correct()
        self.normalize()
        self.writeOut()


if __name__ == '__main__':
    from vencopy.classes.dataParsers import DataParser
    from vencopy.classes.evaluators import Evaluator
    datasetID = 'MiD17'
    pathGlobalConfig = Path.cwd().parent / 'config' / 'globalConfig.yaml'
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathFlexConfig = Path.cwd().parent / 'config' / 'flexConfig.yaml'
    with open(pathFlexConfig) as ipf:
        flexConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathParseConfig = Path.cwd().parent / 'config' / 'parseConfig.yaml'
    with open(pathParseConfig) as ipf:
        parseConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathEvaluatorConfig = Path.cwd().parent / 'config' / 'evaluatorConfig.yaml'
    with open(pathEvaluatorConfig) as ipf:
        evaluatorConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathLocalPathConfig = Path.cwd().parent / 'config' / 'localPathConfig.yaml'
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    os.chdir(localPathConfig['pathAbsolute']['vencoPyRoot'])

    vpData = DataParser(parseConfig=parseConfig, globalConfig=globalConfig, localPathConfig=localPathConfig,
                        loadEncrypted=False)
    vpFlexEst = FlexEstimator(flexConfig=flexConfig, globalConfig=globalConfig, evaluatorConfig=evaluatorConfig,
                              ParseData=vpData, datasetID=datasetID)
    vpFlexEst.run()
    vpEval = Evaluator(globalConfig=globalConfig, evaluatorConfig=evaluatorConfig,
                       parseData=pd.Series(data=vpData, index=[datasetID]))
    vpEval.plotProfiles(flexEstimator=vpFlexEst)
    print(f'Total absolute electricity charged in uncontrolled charging: '
          f'{vpFlexEst.chargeProfilesUncontrolled.sum().sum()} based on MiD17')


__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff'
__contributors__ = 'Fabia Miorelli, Parth Butte'
__email__ = 'niklas.wulff@dlr.de'
__birthdate__ = '23.10.2020'
__status__ = 'prod'  # options are: dev, test, prod
__license__ = 'BSD-3-Clause'


# ----- imports & packages ------
import yaml
import pathlib
import pandas as pd
from vencopy.classes.dataParsers import DataParser
from vencopy.classes.tripDiaryBuilders import TripDiaryBuilder
from vencopy.classes.gridModelers import GridModeler
from vencopy.classes.flexEstimators import FlexEstimator
from vencopy.classes.evaluators import Evaluator


if __name__ == '__main__':
    # Set dataset and config to analyze
    datasetID = 'MiD17'
    # pathLib syntax for windows, max, linux compatibility, see https://realpython.com/python-pathlib/ for an intro
    pathGlobalConfig = pathlib.Path.cwd() / 'config' / 'globalConfig.yaml'
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathLocalPathConfig = pathlib.Path.cwd() / 'config' / 'localPathConfig.yaml'
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathParseConfig = pathlib.Path.cwd() / 'config' / 'parseConfig.yaml'
    with open(pathParseConfig) as ipf:
        parseConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathTripConfig = pathlib.Path.cwd() / 'config' / 'tripConfig.yaml'
    with open(pathTripConfig) as ipf:
        tripConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathGridConfig = pathlib.Path.cwd() / 'config' / 'gridConfig.yaml'
    with open(pathGridConfig) as ipf:
        gridConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathEvaluatorConfig = pathlib.Path.cwd() / 'config' / 'evaluatorConfig.yaml'
    with open(pathEvaluatorConfig) as ipf:
        evaluatorConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathFlexConfig = pathlib.Path.cwd() / 'config' / 'flexConfig.yaml'
    with open(pathFlexConfig) as ipf:
        flexConfig = yaml.load(ipf, Loader=yaml.SafeLoader)

    vpData = DataParser(datasetID=datasetID, parseConfig=parseConfig, globalConfig=globalConfig, localPathConfig=localPathConfig, loadEncrypted=False)

    # Trip distance and purpose diary compositions
    vpTripDiary = TripDiaryBuilder(datasetID=datasetID, tripConfig=tripConfig, globalConfig=globalConfig,
                                   ParseData=vpData, debug=True)

    # Grid model applications
    vpGrid = GridModeler(gridConfig=gridConfig, globalConfig=globalConfig, datasetID=datasetID)
    vpGrid.assignSimpleGridViaPurposes()
    vpGrid.writeOutGridAvailability()

    # Evaluate drive and trip purpose profiles
    vpEval = Evaluator(globalConfig=globalConfig, evaluatorConfig=evaluatorConfig,
                       parseData=pd.Series(data=vpData, index=[datasetID]))
    vpEval.hourlyAggregates = vpEval.calcVariableSpecAggregates(by=['tripStartWeekday'])
    vpEval.plotAggregates()

    # Estimate charging flexibility based on driving profiles and charge connection
    vpFlex = FlexEstimator(flexConfig=flexConfig, globalConfig=globalConfig, evaluatorConfig=evaluatorConfig, datasetID=datasetID, ParseData=vpData)
    vpFlex.baseProfileCalculation()
    vpFlex.filter()
    vpFlex.aggregate()
    vpFlex.correct()
    vpFlex.normalize()
    vpFlex.writeOut()

    vpEval.plotProfiles(flexEstimator=vpFlex)



__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff'
__contributors__ = 'Fabia Miorelli, Parth Butte'
__email__ = 'niklas.wulff@dlr.de'
__birthdate__ = '31.12.2019'
__status__ = 'prod'  # options are: dev, test, prod
__license__ = 'BSD-3-Clause'

import pandas as pd
import yaml
import pathlib


def createFileString(globalConfig: dict, fileKey: str, datasetID: str=None, manualLabel: str = '',
                     filetypeStr: str = 'csv'):
    """
    Generic method used for fileString compilation throughout the VencoPy framework. This method does not write any
    files but just creates the file name including the filetype suffix.

    :param globalConfig: global config file for paths
    :param fileKey: Manual specification of fileKey
    :param datasetID: Manual specification of data set ID e.g. 'MiD17'
    :param manualLabel: Optional manual label to add to filename
    :param filetypeStr: filetype to be written to hard disk
    :return: Full name of file to be written.
    """
    if datasetID is None:
        return f"{globalConfig['files'][fileKey]}_{globalConfig['labels']['runLabel']}_{manualLabel}.{filetypeStr}"
    return f"{globalConfig['files'][datasetID][fileKey]}_{globalConfig['labels']['runLabel']}_{manualLabel}_" \
           f"{datasetID}.{filetypeStr}"


def mergeVariables(data, variableData, variables):
    """
    Global VencoPy function to merge MiD variables to trip distance, purpose or grid connection data.

    :param data: trip diary data as given by tripDiaryBuilder and gridModeler
    :param variableData: Survey data that holds specific variables for merge
    :param variables: Name of variables that will be merged
    :return: The merged data
    """

    variableDataUnique = variableData.loc[~variableData['hhPersonID'].duplicated(), :]
    variables.append('hhPersonID')
    variableDataMerge = variableDataUnique.loc[:, variables].set_index('hhPersonID')
    if 'hhPersonID' not in data.index.names:
        data.set_index('hhPersonID', inplace=True, drop=True)
    mergedData = pd.concat([variableDataMerge, data], axis=1, join='inner')
    mergedData.reset_index(inplace=True)
    return mergedData


def mergeDataToWeightsAndDays(diaryData, ParseData):
    return mergeVariables(data=diaryData, variableData=ParseData.data, variables=['tripStartWeekday', 'tripWeight'])


def calculateWeightedAverage(col, weightCol):
    return sum(col * weightCol) / sum(weightCol)


def writeProfilesToCSV(profileDictOut, globalConfig: dict, singleFile=True, datasetID='MiD17'):
    """
    Function to write VencoPy profiles to either one or five .csv files in the output folder specified in outputFolder.

    :param outputFolder: path to output folder
    :param profileDictOut: Dictionary with profile names in keys and profiles as pd.Series containing a VencoPy
    profile each to be written in value
    :param singleFile: If True, all profiles will be appended and written to one .csv file. If False, five files are
    written
    :param strAdd: String addition for filenames
    :return: None
    """

    if singleFile:
        dataOut = pd.DataFrame(profileDictOut)
        dataOut.to_csv(pathlib.Path(globalConfig['pathRelative']['flexOutput']) /
                       createFileString(globalConfig=globalConfig, fileKey='output',
                                        manualLabel=globalConfig['labels']['technologyLabel'], datasetID=datasetID),
                       header=True)
    else:
        for iName, iProf in profileDictOut.items():
            iProf.to_csv(pathlib.Path(globalConfig['pathRelative']['flexOutput']) /
                         pathlib.Path(f'vencopy_{iName}_{datasetID}.csv'), header=True)



__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff'
__contributors__ = 'Fabia Miorelli, Parth Butte, Ronald Stegen'
__email__ = 'Niklas.Wulff@dlr.de'
__birthdate__ = '30.09.2020'
__status__ = 'prod'  # options are: dev, test, prod
__license__ = 'BSD-3-Clause'


#----- imports & packages ------
from pathlib import Path
import pandas as pd
import yaml
import os
from vencopy.scripts.globalFunctions import createFileString


class GridModeler:
    def __init__(self, gridConfig: dict, globalConfig: dict, datasetID: str ='MiD17'):
        """
        Class for modeling individual vehicle connection options dependent on parking purposes. Configurations on
        charging station availabilities can be parametrized in gridConfig. globalConfig and datasetID are needed for
        reading the input files.

        :param gridConfig: Dictionary containing a key chargingInfrastructureMapping with a sub-dictionary mapping the
        relevant parking purposes to grid availability (true/false). The gridConfig will contain dictionaries with
        probabilistic grid availabilities per parking purpose and rated charging power.
        :param globalConfig: Dictionary with relative paths and filenames. Used for referencing the purpose input file
        :param datasetID: String, used for referencing the purpose input file
        """

        self.inputFileName = createFileString(globalConfig=globalConfig, fileKey='purposesProcessed',
                                              datasetID=datasetID)
        self.inputFilePath = Path(globalConfig['pathRelative']['diaryOutput']) / self.inputFileName
        self.gridDistributions = gridConfig['chargingInfrastructureMappings']
        self.outputFileName = createFileString(globalConfig=globalConfig, fileKey='inputDataPlugProfiles',
                                               datasetID=datasetID)
        self.outputFilePath = Path(globalConfig['pathRelative']['gridOutput']) / self.outputFileName
        self.purposeData = pd.read_csv(self.inputFilePath, keep_default_na=False)
        self.chargeAvailability = None

    def assignSimpleGridViaPurposes(self):
        """
        Method to translate hourly purpose profiles into hourly profiles of true/false giving the charging station
        availability in each hour for each individual vehicle.

        :return: None
        """
        print(f'Starting with charge connection replacement of location purposes')
        self.chargeAvailability = self.purposeData.replace(self.gridDistributions)
        self.chargeAvailability.set_index(['hhPersonID'], inplace=True)
        self.chargeAvailability = (~(self.chargeAvailability != True))
        print('Grid connection assignment complete')

    def writeOutGridAvailability(self):
        """
        Function to write out the boolean charging station availability for each vehicle in each hour to the output
        file path.

        :return: None
        """

        self.chargeAvailability.to_csv(self.outputFilePath)


if __name__ == '__main__':
    pathGlobalConfig = Path.cwd().parent / 'config' / 'globalConfig.yaml'
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathGridConfig = Path.cwd().parent / 'config' / 'gridConfig.yaml'
    with open(pathGridConfig) as ipf:
        gridConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathLocalPathConfig = Path.cwd().parent / 'config' / 'localPathConfig.yaml'
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    os.chdir(localPathConfig['pathAbsolute']['vencoPyRoot'])
    vpg = GridModeler(gridConfig=gridConfig, globalConfig=globalConfig)
    vpg.assignSimpleGridViaPurposes()
    vpg.writeOutGridAvailability()



__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff 31.12.2019'
__contributors__ = 'Fabia Miorelli, Parth Butte'
__email__ = 'Niklas.Wulff@dlr.de'
__birthdate__ = '31.12.2019'
__status__ = 'prod'  # options are: dev, test, prod
__license__ = 'BSD-3-Clause'


import pandas as pd
import numpy as np
from typing import Callable
from pathlib import Path
import yaml
import os
from vencopy.scripts.globalFunctions import createFileString


class TripDiaryBuilder:
    def __init__(self, tripConfig: dict, globalConfig: dict, ParseData, datasetID: str = 'MiD17', debug: bool=False):
        """
        Class to build diaries of trips and parking purposes ("purposes"). Currently a discretization of 1 hour is
        used and tested. Some assumptions are taken in order to build the diaries:
        1. The driven distance is allocated to all hours between start and end time of a trip depending on the time
        share of the respective hour.
        2. The purposes of a parking activity is always determined by the purpose of the previous trip. Purposes are
        taken from the MiD B2 "Wege" (German for trips) data set variable 38 - 'zweck' (German for purpose) and the
        additional purpose "DRIVING" is added.
        3. If an hour comprises multiple purposes (e.g. driving and home), the purpose with the higher share is taken.
        Potential ambiguities in the diaries are later corrected in the gridModel.

        :param tripConfig: In beta release (0.1.0) an empty file
        :param globalConfig: Config holding relative paths, run labels and filenames
        :param ParseData: Class instance of ParseData storing MiD data
        :param datasetID: Can be 'MiD17' or 'MiD08' depending on input data set
        :param debug: Boolean argument. If True, only the first 2000 trips are converted to diaries. This parameter is
        used in the beta since the performance of the current tripDiaryBuilder is quite low.
        """

        self.tripConfig = tripConfig
        self.globalConfig = globalConfig
        self.parsedData = ParseData
        self.tripDataClean = None
        self.tripDistanceDiary = None
        self.tripPurposeDiary = None
        if debug:
            self.tripDataClean = self.calculateConsistentHourlyShares(data=ParseData.data.loc[0:2000, :])
        else:
            self.tripDataClean = self.calculateConsistentHourlyShares(data=ParseData.data)
        self.tripDistanceDiary = self.tripDistanceAllocation(globalConfig)
        self.tripPurposeAllocation()
        self.writeOut(globalConfig=globalConfig, datasetID=datasetID, dataDrive=self.tripDistanceDiary,
                      dataPurpose=self.tripPurposeDiary)

    def tripDuration(self, timestampStart: np.datetime64, timestampEnd: np.datetime64) -> np.datetime64:
        return timestampEnd - timestampStart

    def calcHourShareStart(self, timestampStart: pd.Series, timestampEnd: pd.Series, duration: pd.Series) -> (pd.Series,
                                                                                                           pd.Series):
        """
        :param timestampStart: start time of a trip
        :param timestampEnd:  end time of a trip
        :param duration: duration of a trip
        :return: Returns a data frame of share of individual trip for trips completed in an hour and more w.r.t start time of the trip
        """
        isSameHourTrip = timestampStart.dt.hour == timestampEnd.dt.hour
        shareSameHour = (timestampEnd.dt.minute - timestampStart.dt.minute) / (duration.dt.seconds / 60)
        shareSameHour[duration == pd.Timedelta(0)] = 1  # Set share of first hour to 1 for trips with duration of 0
        share = pd.Series(shareSameHour.where(isSameHourTrip,
                                              (60 - timestampStart.dt.minute) / (duration.dt.seconds / 60)))
        return share.copy(), isSameHourTrip

    def calcHourShareEnd(self, timestampEnd: pd.Series, duration: pd.Series, isSameHourTrip: pd.Series) -> pd.Series:
        """
        :param timestampEnd: end time of a trip
        :param duration: duration of a trip
        :param isSameHourTrip: data frame containing same start time of various trips
        :return: Returns a data frame of share of individual trip for trips completed in an hour and more w.r.t end time of the trip
        """
        share = timestampEnd.dt.minute / (duration.dt.seconds / 60)
        return share.where(~isSameHourTrip, 0)

    def calcDistanceShares(self, data: pd.DataFrame, duration: pd.Series, timestampSt: str,
                           timestampEn: str) -> tuple:
        """
        :param data: list of strings declaring the datasetIDs to be read in
        :param duration: duration of a trip
        :param timestampSt:  start time of a trip
        :param timestampEn:  end time of a trip
        :return: Return a data frame of distance covered by each trip in an hour or more
        """
        shareHourStart, isSameHourTrip = self.calcHourShareStart(timestampStart=data.loc[:, timestampSt],
                                                                 timestampEnd=data.loc[:, timestampEn],
                                                                 duration=duration)
        shareHourEnd = self.calcHourShareEnd(timestampEnd=data.loc[:, timestampEn], duration=duration,
                                             isSameHourTrip=isSameHourTrip)
        return shareHourStart.copy(), shareHourEnd.copy()

    def numberOfFullHours(self, timestampStart: pd.Series, timestampEnd: pd.Series) -> pd.DataFrame:
        """
        Calculates the number of full hours in each trip. E.g. a trip taking 1:42 has 1 full hour and a 30-minute trip
        has 0 full hours.

        :param timestampStart:  Start timestamps of trips
        :param timestampEnd: End times of trips
        :return: Returns a data frame of number of full hours of all trips
        """
        timedeltaTrip = timestampEnd - timestampStart
        numberOfHours = timedeltaTrip.apply(lambda x: x.components.hours)
        numberOfDays = timedeltaTrip.apply(lambda x: x.components.days)
        minLeftFirstHour = pd.to_timedelta(60 - timestampStart.dt.minute, unit='m')
        hasFullHourAfterFirstHour = (timedeltaTrip - minLeftFirstHour) >= pd.Timedelta(1, unit='h')
        numberOfHours = numberOfHours.where(hasFullHourAfterFirstHour, other=0)
        return numberOfHours.where(numberOfDays != -1, other=0)

    def calcFullHourTripLength(self, duration: pd.Series, numberOfFullHours: pd.Series, tripLength: pd.Series) -> \
            pd.Series:
        """
        Calculates the share of the full trip hours. E.g. the fullHourTripLength of a trip starting at 1:45 and ending
        at 4:30 is 120 minutes / 165 minutes ~ 0.73.

        :param duration: Series holding durations of all trips
        :param numberOfFullHours: Series holding the number of full hours of all trips
        :param tripLength: Series of trip lengths
        :return: Returns a Series of full hour trip lengths for all trips
        """

        fullHourTripLength = (numberOfFullHours / (duration.dt.seconds / 3600)) * tripLength
        fullHourTripLength.loc[duration == pd.Timedelta(0)] = 0  # set trip length to 0 that would otherwise be NaN
        return fullHourTripLength

    def calcHourlyShares(self, data: pd.DataFrame, ts_st: str, ts_en: str) -> pd.DataFrame:
        """
        Calculates the shares of first hour, the share of full hours and the full hour trip length.

        :param data: Trip data
        :param ts_st: String specifying the trip start column
        :param ts_en: String specifying the trip end column
        :return: data frame consisting additional information regarding share of a trip, number of full hours and lenght of each trip
        """

        duration = self.tripDuration(data.loc[:, ts_st], data.loc[:, ts_en])
        data.loc[:, 'shareStartHour'], data.loc[:, 'shareEndHour'] = self.calcDistanceShares(data, duration, ts_st, ts_en)
        data.loc[:, 'noOfFullHours'] = self.numberOfFullHours(data.loc[:, ts_st], data.loc[:, ts_en])
        data.loc[:, 'fullHourTripLength'] = self.calcFullHourTripLength(duration, data.loc[:, 'noOfFullHours'],
                                                                   data.loc[:, 'tripDistance'])
        return data

    def calculateConsistentHourlyShares(self, data: pd.DataFrame):
        """
        Wrapper around calcHourlyShares also filtering inconsistent trips.

        :param data: Trip data
        :return: Filtered trip data
        """

        print('Calculating hourly shares')
        if not data._is_view:
            data = data.copy()  #FIXME: why is data._is_view False if we get a view
        tripDataWHourlyShares = self.calcHourlyShares(data, ts_st='timestampStart', ts_en='timestampEnd')

        # Filter out implausible hourly share combinations
        return tripDataWHourlyShares.loc[~((tripDataWHourlyShares['shareStartHour'] != 1) &
                                                       (tripDataWHourlyShares['shareEndHour'] == 0) &
                                                       (tripDataWHourlyShares['noOfFullHours'] == 0)), :]

    def initiateHourDataframe(self, indexCol, nHours: int) -> pd.DataFrame:
        """
        Sets up an empty dataframe to be filled with hourly data.

        :param indexCol: List of column names
        :param nHours: integer giving the number of columns that should be added to the dataframe
        :return: data frame with columns given and nHours additional columns appended with 0s
        """
        return pd.DataFrame(index=indexCol, columns=range(nHours))

    def fillDataframe(self, hourlyArray: pd.DataFrame, fillFunction) -> pd.DataFrame:
        return hourlyArray.apply(fillFunction, axis=1)

    def mergeTrips(self, tripData: pd.DataFrame) -> pd.DataFrame:
        """
        Merge multiple individual hourly trip distances into one diary consisting of multiple trips

        :param tripData: Input trip data with hourly distances of all hourly trips
        :return: Merged trip distance diaries
        """
        dataDay = tripData.groupby(['hhPersonID']).sum()
        dataDay = dataDay.drop('tripID', axis=1)
        return dataDay

    def initiateColRange(self, row: pd.Series):
        """
        Returns a range object with start and end hour as limits

        :param row: A trip observation
        :return:
        """
        if row['tripStartHour'] + 1 < row['tripEndHour']:
            return range(row['tripStartHour'] + 1, row['tripEndHour'])  # The hour of arrival (tripEndHour) will
            # not be indexed further below but is part of the range() object
        else:
            return None

    def tripDistanceAllocation(self, globalConfig : dict) -> pd.DataFrame:
        """
        Wrapper function for the conversion of trip distance values (in a column) to hourly trip distance diaries.

        :param globalConfig: Dictionary holding relative paths, filenames and run labels
        :return: Trip distance diary as a pd.DataFrame
        """
        print('Trip distance diary setup starting')
        self.formatDF = self.initiateHourDataframe(indexCol=self.tripDataClean.index, nHours=globalConfig['numberOfHours'])
        fillHourValues = FillHourValues(data=self.tripDataClean, rangeFunction=self.initiateColRange)
        driveDataTrips = self.fillDataframe(self.formatDF, fillFunction=fillHourValues)
        driveDataTrips.loc[:, ['hhPersonID', 'tripID']] = pd.DataFrame(self.tripDataClean.loc[:, ['hhPersonID',
                                                                                                  'tripID']])
        driveDataTrips = driveDataTrips.astype({'hhPersonID': int, 'tripID': int})
        print('Finished trip distance diary setup')
        return self.mergeTrips(driveDataTrips)

    def assignDriving(self, driveData: pd.DataFrame) -> pd.DataFrame:
        """
        Assign hours where driveData != 0/NA to 'driving'

        :param driveData: driving data
        :return: Returns driving data with 'driving' instead of hours having 0/NA
        """
        locationData = driveData.copy()
        locationData = locationData.where(locationData == 0, other='DRIVING')
        return locationData

    def determinePurposeStartHour(self, departure: np.datetime64, arrival: np.datetime64) -> int:
        """
        Determines the start hour of a parking activity depending on previous trip end time and next trip start time

        :param departure: Start time of next trip after parking
        :param arrival: End time of previous trip
        :return: Returns start hour of a parking activity
        """

        if departure.hour == arrival.hour:
            if arrival.minute >= 30:  # Cases 3, 4, 5
                startHour = departure.hour + 1  # Cases 3,5
            else:  # arrival.minute < 30:
                startHour = departure.hour  # Case 4
        else:  # inter-hour trip
            if arrival.minute <= 30:
                startHour = arrival.hour  # Cases 1a and b
            else:  # arrival.minute > 30:
                startHour = arrival.hour + 1  # Cases 2a and b
        return startHour

    def fillDayPurposes(self, tripData: pd.DataFrame, purposeDataDays: pd.DataFrame) -> pd.DataFrame:  # FixMe: Ask Ben for performance improvements
        """
        Main purpose diary builder function. Root of low performance of tripDiaryBuilder. Will be improved in future
        releases.

        :param tripData: data frame holding all the information about individual trip
        :param purposeDataDays: DataFrame with 24 (hour) columns holding 0s and 'DRIVING' for trip hours (for hours
        where majority of time is driving)
        :return: Returns a data frame of individual trip with it's hourly activity or parking purpose
        """
        hpID = str()
        maxWID = int()
        maxHour = len(purposeDataDays.columns)

        # # uniques = tripData['hhPersonID'].unique()
        #
        # for iSubData in tripData.groupby('hhPersonID'):
        #
        #     currentPerson = tripData['hhPersonID'] == hpID
        #     allWIDs = tripData.loc[currentPerson, 'tripID']  # FIXME perf
        #     minWID = allWIDs.min()  # FIXME perf
        #     maxWID = allWIDs.max()  # FIXME perf
        #
        #     isFirstTripID = iSubData['tripID'] == 1
        #
        #     isBelowHalfHour = iSubData['timestampStart'].dt.minute <= 30
        #
        #     isMaxWID = iSubData['tripID'] == maxWID
        #
        #     isMinWID = iSubData['tripID'] == minWID
        #
        #     arrivalEqualsDeparture = iSubData['timestampEnd'].dt.hour == iSubData['timestampStart'].dt.hour
        #
        #     arrivalIsBelowHalfHour = iSubData['timestampStart'].dt.hour <= 30

        # Solution 1: use enumerate in order to get rowNumber instead of index and then .iloc below
        # Solution 2: Rename columns
        for idx, iRow in tripData.iterrows():
            isSameHPID = hpID == iRow['hhPersonID']
            if not isSameHPID:
                hpID = iRow['hhPersonID']
                allWIDs = tripData.loc[tripData['hhPersonID'] == hpID, 'tripID']  # FIXME perf
                minWID = allWIDs.min()  # FIXME perf
                maxWID = allWIDs.max()  # FIXME perf

            if iRow['tripID'] == 1:  # Differentiate if trip starts in first half hour or not

                if iRow['timestampStart'].minute <= 30:
                    # purposeDataDays.loc[hpID, 0:iRow['tripStartHour']] = 'HOME'  # FIXME perf
                    purposeDataDays.loc[hpID, range(0, iRow['tripStartHour'])] = 'HOME'
                else:
                    purposeDataDays.loc[hpID, range(0, iRow['tripStartHour'] + 1)] = 'HOME'
                if iRow['tripID'] == maxWID:
                    if iRow['timestampEnd'].minute <= 30:
                        purposeDataDays.loc[hpID, range(iRow['tripEndHour'], maxHour)] = 'HOME'
                    else:
                        purposeDataDays.loc[hpID, range(iRow['tripEndHour']+1, maxHour)] = 'HOME'
            elif iRow['tripID'] == minWID:
                if iRow['timestampStart'].minute <= 30:
                    purposeDataDays.loc[hpID, range(0, iRow['tripStartHour'])] = 'HOME'
                else:
                    purposeDataDays.loc[hpID, range(0, iRow['tripStartHour'] + 1)] = 'HOME'
                if iRow['tripID'] == maxWID:
                    purposeDataDays.loc[hpID, range(iRow['tripEndHour'] + 1, maxHour)] = 'HOME'
            else:
                purposeHourStart = self.determinePurposeStartHour(tripData.loc[idxOld, 'timestampStart'],
                                                             tripData.loc[idxOld, 'timestampEnd'])  # FIXME perf?
                if iRow['timestampStart'].minute <= 30:
                    hoursBetween = range(purposeHourStart,
                                         iRow['tripStartHour'])  # FIXME: case differentiation on arrival hour
                else:
                    hoursBetween = range(purposeHourStart,
                                         iRow['tripStartHour'] + 1)
                purposeDataDays.loc[hpID, hoursBetween] = tripData.loc[idxOld, 'purposeStr']

                # NEW PERFORMANCE IMPROVED SNIPPET
                # if iRow['timestampStart'].minute <= 30:
                #     purposeDataDays.loc[hpID, range(purposeHourStart,
                #                                     iRow['tripStartHour'])] = tripData.loc[idxOld, 'purposeStr']
                #     # hoursBetween = range(purposeHourStart,
                #     #                      iRow['tripStartHour'])  # FIXME: case differentiation on arrival hour
                # else:
                #     # hoursBetween = range(purposeHourStart,
                #     #                      iRow['tripStartHour'] + 1)
                #     purposeDataDays.loc[hpID, purposeHourStart:iRow['tripStartHour'] + 1] = tripData.loc[idxOld, 'purposeStr']
                if iRow['tripID'] == maxWID:
                    if iRow['timestampEnd'].minute <= 30:
                        purposeDataDays.loc[hpID, range(iRow['tripEndHour'], maxHour)] = 'HOME'
                    else:
                        purposeDataDays.loc[hpID, range(iRow['tripEndHour'] + 1, maxHour)] = 'HOME'
            idxOld = idx
        return purposeDataDays

    def tripPurposeAllocation(self):
        """
        Wrapper function for trip purpose allocation. Falsely non-allocated parking purposes are replaced by "HOME".

        :return: None
        """
        print('Starting trip purpose diary setup')
        tripPurposesDriving = self.assignDriving(self.tripDistanceDiary)
        self.tripPurposeDiary = self.fillDayPurposes(tripData=self.tripDataClean, purposeDataDays=tripPurposesDriving)
        self.tripPurposeDiary.replace({'0.0': 'HOME'})  # Replace remaining non-allocated purposes with HOME
        print('Finished purpose replacements')
        print(f'There are {len(self.tripPurposeDiary)} daily trip diaries.')

    # improved purpose allocation approach
    def mapHHPIDToTripID(self, tripData):
        idCols = self.tripDataClean.loc[:, ['hhPersonID', 'tripID']]
        idCols.loc['nextTripID'] = idCols['tripID'].shift(-1, fill_value=0)
        tripDict = dict.fromkeys(set(idCols['hhPersonID']))
        for ihhpID in tripDict.keys():
            tripDict[ihhpID] = set(idCols.loc[idCols['hhPersonID'] == ihhpID, 'tripID'])
        return tripDict

    def writeOut(self, globalConfig:dict, dataDrive: pd.DataFrame, dataPurpose: pd.DataFrame, datasetID: str = 'MiD17'):
        """
        General writeout utility for tripDiaries

        :param globalConfig: global config storing relative paths, filenames and run labels
        :param dataDrive: Driving distance diary for each survey participant
        :param dataPurpose: Parking purpose diary for each survey participant
        :param datasetID: ID used for filenames
        :return: None
        """
        dataDrive.to_csv(Path(globalConfig['pathRelative']['diaryOutput']) /
                         createFileString(globalConfig=globalConfig, fileKey='inputDataDriveProfiles',
                                          datasetID=datasetID),
                         na_rep=0)
        dataPurpose.to_csv(Path(globalConfig['pathRelative']['diaryOutput']) /
                          createFileString(globalConfig=globalConfig, fileKey='purposesProcessed', datasetID=datasetID))
        print(f"Drive data and trip purposes written to files "
              f"{createFileString(globalConfig=globalConfig, fileKey='inputDataDriveProfiles', datasetID=datasetID)} "
              f"and {createFileString(globalConfig=globalConfig, fileKey='purposesProcessed', datasetID=datasetID)}")


class FillHourValues:
    def __init__(self, data, rangeFunction):
        self.startHour = data['tripStartHour']
        self.distanceStartHour = data['shareStartHour'] * data['tripDistance']
        self.endHour = data['tripEndHour']
        self.distanceEndHour = data['shareEndHour'] * data['tripDistance']
        self.fullHourCols = data.apply(rangeFunction, axis=1)
        self.fullHourRange = data['fullHourTripLength']

    def __call__(self, row):
        idx = row.name
        row[self.startHour[idx]] = self.distanceStartHour[idx]
        if self.endHour[idx] != self.startHour[idx]:
            row[self.endHour[idx]] = self.distanceEndHour[idx]
        if isinstance(self.fullHourCols[idx], range):
            row[self.fullHourCols[idx]] = self.fullHourRange[idx]
        return row


if __name__ == '__main__':
    from vencopy.classes.dataParsers import DataParser
    pathGlobalConfig = Path.cwd().parent / 'config' / 'globalConfig.yaml'  # pathLib syntax for windows, max, linux compatibility, see https://realpython.com/python-pathlib/ for an intro
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathParseConfig = Path.cwd().parent / 'config' / 'parseConfig.yaml'
    with open(pathParseConfig) as ipf:
        parseConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathTripConfig = Path.cwd().parent / 'config' / 'tripConfig.yaml'
    with open(pathTripConfig) as ipf:
        tripConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathLocalPathConfig = Path.cwd().parent / 'config' / 'localPathConfig.yaml'
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    os.chdir(localPathConfig['pathAbsolute']['vencoPyRoot'])
    vpData = DataParser(parseConfig=parseConfig, globalConfig=globalConfig, localPathConfig=localPathConfig,
                        loadEncrypted=False)
    vpDiary = TripDiaryBuilder(tripConfig=tripConfig, globalConfig=globalConfig, ParseData=vpData)


"""
VencoPy.

Vehicle Energy Consumption in Python. Electric vehicle fleet load shifting simulations.
"""

__version__ = "0.1.0"
__author__ = 'Niklas Wulff'
__credits__ = 'German Aerospace Center (DLR)'



__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff'
__contributors__ = 'Fabia Miorelli, Parth Butte'
__email__ = 'Niklas.Wulff@dlr.de'
__birthdate__ = '21.09.2020'
__status__ = 'test'  # options are: dev, test, prod
__license__ = 'BSD-3-Clause'

import yaml
import os
import pathlib
import pandas as pd
import seaborn as sns
from pathlib import Path
import matplotlib.pyplot as plt
from vencopy.scripts.globalFunctions import createFileString, calculateWeightedAverage, mergeDataToWeightsAndDays, \
    writeProfilesToCSV
# from classes.flexEstimators import FlexEstimator


class Evaluator:
    def __init__(self, globalConfig:dict, evaluatorConfig: dict, parseData: pd.Series = None, weightPlot=True):
        """
        CURRENTLY IN SEMI-PROUDCTION MODE. Some interfaces may only apply to specific cases.
        Overall evaluation class for assessing vencopy mobility and charging profiles.

        :param globalConfig: global config instances for paths
        :param evaluatorConfig: evaluator config mainly for plot properties
        :param label: String for saving plots to hard disk
        :param parseData: Series with instances of VencoPy class ParseData and keys specifying the name of the
            respective class
        :param weightPlot: If True, profiles are weighted for plotting
        """

        self.globalConfig = globalConfig
        self.evaluatorConfig = evaluatorConfig
        self.weightPlot = weightPlot
        self.normPlotting = True
        self.dailyMileageGermany2008 = 3.080e9  # pkm/d
        self.dailyMileageGermany2017 = 3.214e9  # pkm/d
        self.hourVec = [str(i) for i in range(0, globalConfig['numberOfHours'])]
        if parseData is not None:
            self.datasetIDs = list(parseData.index)
            self.parseData = parseData
            self.inputDataRaw = self.readInData(['inputDataDriveProfiles'], self.datasetIDs)
            self.inputData = pd.Series(dtype=object)
            self.mergeDaysAndWeights()
            self.data = pd.Series(dtype=object)
            self.reindexData()
            self.dataStacked = None
            self.aggregateIDDict = self.setupAggDict()
            self.hourlyAggregates = self.aggregateAcrossTrips()

        print('Evaluator initialization complete')

    def readInData(self, fileKeys: list, datasets: list) -> pd.Series:
        """
        Generic read-in function for mobility datasetIDs. This serves as interface between the daily trip distance
        and purpose calculation and the class Evaluator.

        :param fileKeys: List of VencoPy-internal names for the filekeys to read in
        :param datasets: list of strings declaring the datasetIDs to be read in
        :return: a named pd.Series of all datasetIDs with the given filekey_datasets as identifiers
        """

        ret = pd.Series(dtype=object)
        for iFileKey in fileKeys:
            for iDat in datasets:
                dataIn = pd.read_csv(pathlib.Path(self.globalConfig['pathRelative']['diaryOutput']) /
                                     createFileString(globalConfig=self.globalConfig, fileKey=iFileKey,
                                                      datasetID=iDat), dtype={'hhPersonID': int},
                                     # index_col=['hhPersonID', 'tripStartWeekday'])
                                     index_col=['hhPersonID'])
                ret[iDat] = dataIn
        return ret

    def mergeDaysAndWeights(self):
        """
        Method to merge mobility data to respective trip weights and days.

        :return: None
        """
        for iDat in self.datasetIDs:
            self.inputData[iDat] = mergeDataToWeightsAndDays(self.inputDataRaw[iDat], self.parseData[iDat])
            self.inputData[iDat].dropna()

    # DEPRECATED WILL BE DELETED ON NEXT RELEASE
    def assignWeight(self, datasetIDs: list):
        """
        Reformatting function

        :param datasetIDs:
        :return:
        """

        ret = pd.Series(dtype=object)
        for iDat in datasetIDs:
            weightData = mergeDataToWeightsAndDays(self.inputData[iDat], self.parseData[iDat])
            weights = weightData.loc[:, ['hhPersonID', 'tripStartWeekday', 'tripWeight']]
            weights = weights.convert_dtypes()
            ret[iDat] = weights.set_index(['hhPersonID'], drop=True)
        return ret

    def calculateMobilityQuota(self, dataset: str) -> None:
        """
        Calculates the number of survey days where mobiity occured.

        :param dataset: name of dataset
        :return: Scalar, the ratio of mobile days to total days
        """
        dataKey = f'inputDataDriveProfiles_{dataset}'
        if not dataKey in self.inputData.keys():
            assert 'Specified dataset was not read in during Evaluator initialization'
        dat = self.inputData[dataKey]
        isNoTrip = dat == 0
        isNoTrip.apply(any, axis=1)

    def reindexData(self):
        """
        Formatting function to set index of profiles with weekday and weight, in preparation for plotting

        :return: None
        """

        for iDat in self.datasetIDs:
            self.data[iDat] = self.inputData[iDat].set_index(['hhPersonID', 'tripStartWeekday'], drop=True)
            self.data[iDat].dropna(inplace=True)

    # Maybe not needed at all?
    def stackData(self):
        """
        Function to rearrange and rename data for plotting

        :return: None
        """

        ret = {}
        for idx, iDat in self.inputData.items():
            iDatRaw = iDat.drop(columns=['hhPersonID']).set_index('tripStartWeekday',
                                                                              append=True).stack()  # 'tripWeight', 'tripScaleFactor'
            iDat = iDatRaw.reset_index([1, 2])
            iDat.columns = ['Day', 'Hour', 'Value']
            ret[idx] = iDat
        return ret

    def setupAggDict(self):
        """
        Setup lookup-dictionaries for plotting.

        :return: Updated dictionary of data set IDs
        """

        IDDict = {'sum': None, 'mean': None, 'wMean': None}
        IDDict['sum'] = [f'{iDatID}_sum' for iDatID in self.datasetIDs]
        IDDict['mean'] = [f'{iDatID}_mean' for iDatID in self.datasetIDs]
        IDDict['wMean'] = [f'{iDatID}_wMean' for iDatID in self.datasetIDs]
        return IDDict

    def aggregateAcrossTrips(self):
        """
        Aggregate trip distances across all trops for specific hours. Aggregation is carried out threefold: Summation,
        simple average and weighted average if self.weightPlot is True.

        :return: DataFrame with a column for each aggregated time series and a row for each hour in the original data
            set
        """

        ret = pd.DataFrame()
        for iDatID, iDat in self.inputData.items():
            ret.loc[:, f'{iDatID}_sum'] = iDat.loc[:, self.hourVec].sum(axis=0)
            ret.loc[:, f'{iDatID}_mean'] = iDat.loc[:, self.hourVec].mean(axis=0)
            if self.weightPlot:
                ret.loc[:, f'{iDatID}_wMean'] = iDat.loc[:, self.hourVec].apply(calculateWeightedAverage,
                                                                                args=[iDat.loc[:, 'tripWeight']])
        return ret

    def calcVariableSpecAggregates(self, by: list):
        """
        Aggregation method that calculated aggregates such in aggregateAcrossTrips but differentiating one additional
        criterium given in by. The method is used to calculate weekday specific aggregates.

        :param by: List of column names to add differentiation. Currently only tested for tripStartWeekday
        :return: Pandas DataFrame with three columns each one for summed, average and weighted average and a multiindex
            differentiating between hours and one additional variable specificied in by.
        """

        ret = pd.DataFrame()
        for iDatID, iDat in self.data.items():
            if not all([iBy in iDat.index.names for iBy in by]):
                raise Exception('At least one required variable name is not in index names. Aborting')
            ret.loc[:, f'{iDatID}_sum'] = iDat.loc[:, self.hourVec].groupby(level=by).sum().stack()  # needed .stack() here?
            ret.loc[:, f'{iDatID}_mean'] = iDat.loc[:, self.hourVec].groupby(level=by).mean().stack()
            if self.weightPlot:
                # ret.loc[:, f'{iDatID}_wMean'] = iDat.loc[:,
                #                                 self.hourVec].groupby(level=by).apply(calculateWeightedAverage,
                #                                                     weightCol=[iDat.loc[:, 'tripWeight']]).stack()
                ret.loc[:, f'{iDatID}_wMean'] = self.calcWeightedTripValues(dataIn=iDat, idxLvl=by[0])
        return ret

    def calcWeightedTripValues(self, dataIn, idxLvl):
        """
        Function to calculate weighted trip values with an additional differentiating column given in idxLvl.

        :param dataIn: Input data for aggregation
        :param idxLvl: Index level specifying additional differentiation criterium
        :return:
        """
        vars = set(dataIn.index.get_level_values(idxLvl))
        ret = pd.DataFrame(index=self.hourVec, columns=vars)
        data = dataIn.loc[:, self.hourVec].reset_index(level=idxLvl)
        weights = dataIn.loc[:, 'tripWeight'].reset_index(level=idxLvl)
        for iVar in vars:
            dataSlice = data.loc[data.loc[:, idxLvl] == iVar, self.hourVec]
            weightSlice = weights.loc[data.loc[:, idxLvl] == iVar, 'tripWeight']
            ret.loc[:, iVar] = dataSlice.apply(calculateWeightedAverage, weightCol=weightSlice)
        ret = ret.stack()
        ret.index = ret.index.swaplevel(0, 1)
        return ret

    def sortData(self, data):
        """
        Method used for plotting to order index values

        :param data: Pandas Dataframe with two indices
        """
        data.index = data.index.swaplevel(0, 1)
        return data.sort_index()

    def plotAggregates(self):
        """
        Plotting method for plotting aggregates that are differentiated with one additional variable. In the current
        implementation both summed and averaged values are plotted. Arguments show and write in the evaluator config
        determine if plots are shown to the user and/or written to hard disk. X axis ticks are currently manually
        configured to fit weekday and hour differentiation for one week.

        :return: None
        """

        self.hourlyAggregates = self.hourlyAggregates.swaplevel(0, 1)
        # Plotting aggregates
        fig, ax = plt.subplots(2, 1)
        plt.tick_params(labelsize=self.evaluatorConfig['plotConfig']['plotRCParameters']['font.size'])
        meanCols = self.aggregateIDDict['mean']
        meanCols.extend(self.aggregateIDDict['wMean'])
        self.hourlyAggregates.loc[:, self.aggregateIDDict['sum']].plot.line(ax=ax[0])
        self.hourlyAggregates.loc[:, meanCols].plot.line(ax=ax[1])
        xRange = range(0, len(self.hourlyAggregates) + 1, self.evaluatorConfig['plotConfig']['xAxis']['xTickSteps'])
        xLabels = [f'{iDay}\n{str(iTime)}:00' for iDay in self.evaluatorConfig['plotConfig']['xAxis']['weekdays']
                   for iTime in self.evaluatorConfig['plotConfig']['xAxis']['hours']]
        # xLabels = [f'{str(iTime)}:00' for iTime in self.evaluatorConfig['plotConfig']['xAxis']['hours']]
        ax[1].set_xticks(xRange[:-1])
        ax[1].set_xticklabels(xLabels, fontsize=self.evaluatorConfig['plotConfig']['xAxis']['ticklabelsize'])
        ax[0].ticklabel_format(axis='y', style='sci', scilimits=(0, 0), useMathText=True)
        ax[0].set_xticks(xRange[:-1])
        ax[0].set_xticklabels('')
        ax[0].set_xlabel('')
        ax[0].set_ylabel('Sum of all trips \n in sample in km',
                         fontsize=self.evaluatorConfig['plotConfig']['yAxis']['ticklabelsize'])
        ax[1].set_xlabel("Hour", fontsize=self.evaluatorConfig['plotConfig']['plotRCParameters']['axes.labelsize'])
        ax[1].set_ylabel("Average of hourly \n trips in km",
                         fontsize=self.evaluatorConfig['plotConfig']['plotRCParameters']['axes.labelsize'])
        ax[0].tick_params(axis='y', labelsize=self.evaluatorConfig['plotConfig']['yAxis']['ticklabelsize'])
        ax[0].legend()
        ax[1].legend()

        if self.evaluatorConfig['plotConfig']['show']:
            plt.show()
        if self.evaluatorConfig['plotConfig']['save']:
            fileName = createFileString(globalConfig=self.globalConfig, fileKey='aggPlotName', manualLabel=self.globalConfig['labels']['runLabel'],
                                        filetypeStr='svg')
            fig.savefig(Path(self.globalConfig['pathRelative']['evalOutput']) / fileName, bbox_inches='tight')

    def linePlot(self, profileDict, pathOutput, flexEstimator, show=True, write=True, ylabel='Normalized profiles',
                 ylim=None, filename=''):
        """
        Basic line plot functionality

        :param profileDict: Dictionary specifiying which profiles should be plotted together in one plot
        :param pathOutput: Path to write Figure to
        :param flexEstimator: Instance of VencoPy class FlexEstimator for plotting of charging profiles
        :param show: If True, Figure is displayed during runtime
        :param write: If True, Figrue is written to hard disk
        :param ylabel: Label for y Axis
        :param ylim: Manually set y limit
        :param filename: Manually set name of file written to hard disk
        :return: None
        """

        plt.rcParams.update(self.evaluatorConfig['plotConfig']['plotRCParameters'])  # set plot layout
        fig, ax = plt.subplots()
        plt.tick_params(labelsize=self.evaluatorConfig['plotConfig']['plotRCParameters']['font.size'])
        for iKey, iVal in profileDict.items():
            if isinstance(iVal.index, pd.MultiIndex):
                iVal = self.sortData(iVal)
                sns.lineplot(x=range(iVal.index.size), y=iVal, label=iKey, sort=False)
            else:
                sns.lineplot(x=iVal.index, y=iVal, label=iKey, sort=False)
        xRange = range(0, len(profileDict[list(profileDict)[0]]) + 1,
                       self.evaluatorConfig['plotConfig']['xAxis']['xTickSteps'])
        xLabels = [f'{iDay}\n{str(iTime)}:00' for iDay in self.evaluatorConfig['plotConfig']['xAxis']['weekdays']
                   for iTime in self.evaluatorConfig['plotConfig']['xAxis']['hours']]

        # Labeling for 24 hour plots
        # xLabels = [f'{str(iTime)}:00' for iTime in self.evaluatorConfig['plotConfig']['xAxis']['hours']]
        ax.set_xticks(xRange[:-1])
        ax.set_xticklabels(xLabels, fontsize=self.evaluatorConfig['plotConfig']['xAxis']['ticklabelsize'])
        if ylim:
            ax.set_ylim(bottom=0, top=ylim)
        ax.set_xlabel('Weekday and Hour',
                      fontsize=self.evaluatorConfig['plotConfig']['plotRCParameters']['axes.labelsize'])
        ax.set_ylabel(ylabel, fontsize=self.evaluatorConfig['plotConfig']['plotRCParameters']['axes.labelsize'])
        plt.legend(loc='upper center')
        plt.tight_layout()
        filePlot = pathOutput / Path(
            createFileString(globalConfig=self.globalConfig, datasetID=flexEstimator.datasetID, fileKey='flexPlotName',
                             manualLabel=filename, filetypeStr='svg'))
        if show:
            plt.show()
        if write:
            fig.savefig(filePlot)

    def separateLinePlots(self, profileDictList: list, flexEstimator, show=True,
                          write=True, ylabel=[], ylim=[], filenames=[]):
        """
        Wrapper function to draw and write multiple plots using linePlot().

        :param profileDictList: List of dictionaries. Each dictionary specifies one plot drawn.
        :param flexEstimator: Instance of VencoPy class FlexEstimator for data set ID
        :param show: If True, plots are shown to the user during runtime
        :param write: If True, plots are written to hard drive
        :param ylabel: List of ylabels. Has to be of same length as profileDictList
        :param ylim: List of ylimits. Has to be of same length as profileDictList
        :param filenames: Name of file to be written to hard drive
        :return: None
        """
        if ylim:
            for iDict, iYLabel, iYLim, iName in zip(profileDictList, ylabel, ylim, filenames):
                writeProfilesToCSV(profileDictOut=iDict, globalConfig=self.globalConfig, singleFile=False,
                                   datasetID=flexEstimator.datasetID)
                self.linePlot(iDict, pathOutput=Path(self.globalConfig['pathRelative']['evalOutput']),
                              flexEstimator=flexEstimator, show=show, write=write, ylabel=iYLabel, ylim=iYLim,
                              filename=iName)
        else:
            for iDict, iYLabel, iName in zip(profileDictList, ylabel, filenames):
                writeProfilesToCSV(profileDictOut=iDict, globalConfig=self.globalConfig, singleFile=False,
                                   datasetID=flexEstimator.datasetID)
                self.linePlot(iDict, pathOutput=Path(self.globalConfig['pathRelative']['evalOutput']),
                              flexEstimator=flexEstimator, show=show, write=write, ylabel=iYLabel,
                              filename=iName)

    def plotProfiles(self, flexEstimator, profileDictList: dict = None, yLabels: list = None, yLimits: list = None,
                     filenames: list = None):
        """
        Wrapper function to plot both one Figure with all resulting output profiles and separate Figures for flow,
        connection and state profiles after VencoPy flexibility estimation.

        :param flexEstimator: Instance of VencoPy class FlexEstimator
        :param profileDictList: List of diciontaries with keys specifying the profile name and value holding a pandas
        dataframe with a multiindex series
        :param yLabels: list of y axis labels. Has to be of same length as profileDictList
        :param yLimitsL List of y axis limits. Has to be of same length as profileDictList
        :param filenames: List of filenames to write the plots to. Has to be of same length as profileDictList
        :return: None
        """

        if profileDictList is None:
            # Separately plot flow and state profiles
            profileDictConnectionShare = dict(gridConnectionShare=flexEstimator.plugProfilesWAggVar)

            # profileDictFlowsNorm = dict(uncontrolledCharging=flexEstimator.chargeProfilesUncontrolledCorr,
            #                             electricityDemandDriving=flexEstimator.electricPowerProfilesCorr,
            #                             gridConnectionShare=flexEstimator.plugProfilesAgg)
            profileDictFlowsAbs = dict(uncontrolledCharging=flexEstimator.chargeProfilesUncontrolledWAggVar,
                                       electricityDemandDriving=flexEstimator.electricPowerProfilesWAggVar)

            # profileDictStateNorm = dict(socMax=flexEstimator.socMaxNorm, socMin=flexEstimator.socMinNorm)
            profileDictStateAbs = dict(socMax=flexEstimator.socMaxVar, socMin=flexEstimator.socMinVar)

            profileDictList = [profileDictConnectionShare, profileDictFlowsAbs, profileDictStateAbs]
            yLabels = ['Average EV connection share', 'Average EV flow in kW', 'Average EV SOC in kWh']
            filenames = [flexEstimator.datasetID + '_connection', flexEstimator.datasetID + '_flows',
                                          flexEstimator.datasetID + '_state']
        self.separateLinePlots(profileDictList, show=True, write=True, flexEstimator=flexEstimator,
                               ylabel=yLabels, filenames=filenames)

    def compareProfiles(self, compareTo):
        """
        EXPERIMENTAL STATE. Deprecated method to compare profiles of MiD08 and MiD17. Currently not in production and
        not tested.
        """

        # if not isinstance(compareTo, FlexEstimator):
        #     raise('Argument to compare to is not a class instance of FlexEstimator')

        profileList = [
                       # 'plugProfilesAgg', 'plugProfilesWAgg', 'chargeProfilesUncontrolledAgg',
                       # 'chargeProfilesUncontrolledWAgg', 'electricPowerProfilesAgg', 'electricPowerProfilesWAgg',
                       # 'plugProfilesWAggVar', 'electricPowerProfilesWAggVar', 'chargeProfilesUncontrolledWAggVar'
                       # 'auxFuelDemandProfilesWAggVar',
                       ]

        profileDictList = self.compileDictList(compareTo=compareTo, profileNameList=profileList)
        SOCDataWeek = {'MiD08_SOCmin': self.SOCMinVar,
                       'MiD08_SOCmax': self.SOCMaxVar,
                       'MiD17_SOCmin': compareTo.SOCMinVar,
                       'MiD17_SOCmax': compareTo.SOCMaxVar}

        profileDictList.append(SOCDataWeek)

        self.separateLinePlots(profileDictList, show=self.evaluatorConfig['plotConfig']['show'],
                               write=self.evaluatorConfig['plotConfig']['save'],
                          ylabel=[
                                  # 'Average EV connection share', 'Weighted Average EV connection share',
                                  # 'Uncontrolled charging in kW', 'Weighted Uncontrolled charging in kW',
                                  # 'Electricity consumption for driving in kWh',
                                  # 'Weighted Electricity consumption for driving in kWh',
                                  # 'Weighted average EV fleet connection share',
                                  # 'Electricity consumption for driving in kWh',
                                  # 'Weighted average uncontrolled charging in kW'
                                  # 'auxFuelDemandProfilesWAggVar'
                                  'State of charge in kWh'
                                  ],
                          filenames=[
                                     # '_connection', '_connectionWeighted',
                                     # '_uncCharge', '_uncChargeWeighted',
                                     # '_drain', '_drainWeighted',
                                     # '_plugDiffDay', '_drainDiffDay',
                                     # '_uncChargeDiffDay'
                                     #  '_auxFuelDiffDay',
                                     '_socWeek'
                                     ],
                          ylim=[
                              # 1, 1, 1,
                              # 1, 1, 1
                              # 1, 1, 1
                              # 1
                              50])

    def compileDictList(self, compareTo, profileNameList):
        """
        Helper function for compareProfiles()

        :param compareTo: Dataset to compare to
        :param profileNameList: List of names of files to be written
        :return:
        """

        ret = []
        keys = [self.datasetID, compareTo.datasetID]
        for iProf in profileNameList:
            iDict = self.compileProfileComparisonDict(keys=keys,
                                                      values=[getattr(self, iProf), getattr(compareTo, iProf)])
            ret.append(iDict)
        return ret

    def compileProfileComparisonDict(self, keys: list, values: list):
        return {iKey: iVal for iKey, iVal in zip(keys, values)}

    def compareUncontrolledCharging(self, pathToFile: str, sheetname: str):
        dataIn = pd.read_excel(io=pathToFile, sheet_name=sheetname, header=1, engine='openpyxl')
        data = dataIn.groupby(by=['day']).sum().drop(columns='hour')
        # fix, ax = plt.subplots()
        data.plot()
        plt.show()
        print(data)

    def boxPlot(self, profiles):
        profiles.boxplot()
        plt.show()


if __name__ == '__main__':
    from vencopy.classes.dataParsers import DataParser
    pathGlobalConfig = Path.cwd().parent / 'config' / 'globalConfig.yaml'  # pathLib syntax for windows, max, linux compatibility, see https://realpython.com/python-pathlib/ for an intro
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathParseConfig = Path.cwd().parent / 'config' / 'parseConfig.yaml'
    with open(pathParseConfig) as ipf:
        parseConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathEvaluatorConfig = Path.cwd().parent / 'config' / 'evaluatorConfig.yaml'
    with open(pathEvaluatorConfig) as ipf:
        evaluatorConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathLocalPathConfig = Path.cwd().parent / 'config' / 'localPathConfig.yaml'
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    os.chdir(localPathConfig['pathAbsolute']['vencoPyRoot'])
    parseDataAll = pd.Series(dtype=object)
    parseDataAll['MiD08'] = DataParser(datasetID='MiD08', parseConfig=parseConfig, globalConfig=globalConfig,
                                       localPathConfig=localPathConfig, loadEncrypted=False)
    parseDataAll['MiD17'] = DataParser(datasetID='MiD17', parseConfig=parseConfig, globalConfig=globalConfig,
                                       localPathConfig=localPathConfig, loadEncrypted=False)

    vpEval = Evaluator(globalConfig=globalConfig, evaluatorConfig=evaluatorConfig, parseData=parseDataAll)
    vpEval.hourlyAggregates = vpEval.calcVariableSpecAggregates(by=['tripStartWeekday'])
    vpEval.plotAggregates()

    # vpEval.data = mergeVariables(data=vpEval.inputData['MiD17'].reset_index(),
    #                              variableData=p.data, variables=['tripWeight'])



__version__ = '0.1.0'
__maintainer__ = 'Niklas Wulff 31.12.2019'
__contributors__ = 'Fabia Miorelli, Parth Butte'
__email__ = 'Niklas.Wulff@dlr.de'
__birthdate__ = '31.12.2019'
__status__ = 'prod'  # options are: dev, test, prod

import pprint
import pandas as pd
import numpy as np
import warnings
from pathlib import Path
import yaml
from zipfile import ZipFile


class DataParser:
    def __init__(self, parseConfig: dict, globalConfig: dict, localPathConfig: dict, datasetID: str='MiD17',
                 loadEncrypted=True):
        """
        Basic class for parsing a mobility survey trip data set. Currently the both German travel surveys MiD 2008 and
        MiD 2017 are pre-configured and one of the two can be given (default: MiD 2017).
        The data set can be provided from an encrypted file on a server in which case the link to the ZIP-file as well
        as a link to the file within the ZIP-file have to be supplied in the globalConfig and a password has to be
        supplied in the parseConfig.
        Columns relevant for the EV simulation are selected from the entirety of the data and renamed to VencoPy
        internal variable names given in the dictionary parseConfig['dataVariables'] for the respective survey data set.
        Manually configured exclude, include, greaterThan and smallerThan filters are applied as they are specified in
        parseConfig. For some columns, raw data is transferred to human readable strings and respective columns are
        added. Pandas timestamp columns are synthesized from the given trip start and trip end time information.

        :param parseConfig: A yaml config file holding a dictionary with the keys 'pathRelative'
        :param globalConfig: A yaml config file holding mainly filenames and file labels
        :param localPathConfig: A config thats not synchronized holding local paths
        :param datasetID: Currently, MiD08 and MiD17 are implemented as travel survey data sets
        :param loadEncrypted: If True, load an encrypted ZIP file as specified in parseConfig
        """
        self.datasetID = self.checkDatasetID(datasetID, parseConfig)
        self.parseConfig = parseConfig
        self.globalConfig = globalConfig
        self.rawDataPath = Path(localPathConfig['pathAbsolute'][self.datasetID]) / globalConfig['files'][self.datasetID]['tripsDataRaw']
        self.subDict = {}
        self.rawData = None
        self.data = None
        self.__filterDict = {}
        self.columns = self.compileVariableList()
        self.filterDictNameList = ['include', 'exclude', 'greaterThan', 'smallerThan']
        self.updateFilterDict()
        print('Parsing properties set up')
        if loadEncrypted:
            print(f"Starting to retrieve encrypted data file from "
                  f"{self.globalConfig['pathAbsolute']['encryptedZipfile']}")
            self.loadEncryptedData(pathToZip=Path(self.globalConfig['pathAbsolute']['encryptedZipfile']) /
                                             self.globalConfig['files'][self.datasetID]['encryptedZipFileB2'],
                                   pathInZip=globalConfig['files'][self.datasetID]['tripDataZipFileRaw'])
        else:
            print(f"Starting to retrieve local data file from {self.rawDataPath}")
            self.loadData()
        self.selectColumns()
        self.harmonizeVariables()
        self.convertTypes()
        self.checkFilterDict()
        self.filter()
        self.filterConsistentHours()
        self.addStrColumns()
        self.composeStartAndEndTimestamps()
        print('Parsing completed')

    def updateFilterDict(self) -> None:
        """
        Internal function to parse the filter dictionary of a specified data set from parseConfig.yaml

        :return: None
        """
        self.__filterDict = self.parseConfig['filterDicts'][self.datasetID]
        self.__filterDict = {iKey: iVal for iKey, iVal in self.__filterDict.items() if self.__filterDict[iKey] is not
                             None}

    def checkDatasetID(self, datasetID: str, parseConfig: dict) -> str:
        """
        General check if data set ID is defined in parseConfig.yaml

        :param datasetID: list of strings declaring the datasetIDs to be read in
        :param parseConfig: A yaml config file holding a dictionary with the keys 'pathRelative' and 'pathAbsolute'
        :return: Returns a string value of a mobility data
        """
        availableDatasetIDs = parseConfig['dataVariables']['datasetID']
        assert datasetID in availableDatasetIDs, \
            f'Defined datasetID {datasetID} not specified under dataVariables in parseConfig. Specified datasetIDs ' \
            f'are {availableDatasetIDs}'
        return datasetID

    def compileVariableList(self) -> list:
        """
        Clean up the replacement dictionary of raw data file variable (column) names. This has to be done because some
        variables that may be relevant for the analysis later on are only contained in one raw data set while not
        contained in another one. E.g. if a trip is an intermodal trip was only assessed in the MiD 2017 while it wasn't
        in the MiD 2008. This has to be mirrored by the filter dict for the respective data set.

        :return: List of variables
        """
        listIndex = self.parseConfig['dataVariables']['datasetID'].index(self.datasetID)
        variables = [val[listIndex] if not val[listIndex] == 'NA' else 'NA' for key, val in
                     self.parseConfig['dataVariables'].items()]
        variables.remove(self.datasetID)
        self.removeNA(variables)
        return variables

    def removeNA(self, variables: list):
        """
        Removes all strings that can be capitalized to 'NA' from the list of variables

        :param variables: List of variables of the mobility dataset
        :return: Returns a list with non NA values
        """
        vars = [iVar.upper() for iVar in variables]
        counter = 0
        for idx, iVar in enumerate(vars):
            if iVar == 'NA':
                del variables[idx - counter]
                counter += 1

    def loadData(self):
        """
        Loads data specified in self.rawDataPath and stores it in self.rawData. Raises an exception if a invalid suffix
        is specified in self.rawDataPath. READ IN OF CSV HAS NOT BEEN EXTENSIVELY TESTED BEFORE BETA RELEASE.

        :return: None
        """
        # Future releases: Are potential error messages (.dta not being a stata file even as the ending matches)
        # readable for the user? Should we have a manual error treatment here?
        if self.rawDataPath.suffix == '.dta':
            self.rawData = pd.read_stata(self.rawDataPath, convert_categoricals=False, convert_dates=False,
                                      preserve_dtypes=False)

        # This has not been tested before the beta release
        elif self.rawDataPath.suffix == '.csv':
            self.rawData = pd.read_csv(self.rawDataPath)
        else:
            Exception(f"Data type {self.rawDataPath.suffix} not yet specified. Available types so far are .dta and "
                      f".csv")

        print(f'Finished loading {len(self.rawData)} rows of raw data of type {self.rawDataPath.suffix}')

    def loadEncryptedData(self, pathToZip, pathInZip):
        """
        Since the MiD data sets are only accessible by an extensive data security contract, VencoPy provides the
        possibility to access encrypted zip files. An encryption password has to be given in parseConfig.yaml in order
        to access the encrypted file. Loaded data is stored in self.rawData

        :param pathToZip: path from current working directory to the zip file or absolute path to zipfile
        :param pathInZip: Path to trip data file within the encrypted zipfile
        :return: None
        """
        with ZipFile(pathToZip) as myzip:
            if '.dta' in pathInZip:
                self.rawData = pd.read_stata(myzip.open(pathInZip, pwd=bytes(self.parseConfig['encryptionPW'],
                                                                             encoding='utf-8')),
                                             convert_categoricals=False, convert_dates=False, preserve_dtypes=False)
            else:  # if '.csv' in pathInZip:
                self.rawData = pd.read_csv(myzip.open(pathInZip, pwd=bytes(self.parseConfig['encryptionPW'],
                                                                           encoding='utf-8')), sep=';', decimal=',')

        print(f'Finished loading {len(self.rawData)} rows of raw data of type {self.rawDataPath.suffix}')

    def selectColumns(self):
        """
        Function to filter the rawData for only relevant columns as specified by parseConfig and cleaned in
        self.compileVariablesList(). Stores the subset of data in self.data

        :return: None
        """

        self.data = self.rawData.loc[:, self.columns]
        
    def harmonizeVariables(self):
        """
        Harmonizes the input data variables to match internal VencoPy names given as specified in the mapping in
        parseConfig['dataVariables']. So far mappings for MiD08 and MiD17 are given. Since the MiD08 doesn't provide
        a combined household and person unique identifier, it is synthesized of the both IDs.

        :return: None
        """
        replacementDict = self.createReplacementDict(self.datasetID, self.parseConfig['dataVariables'])
        dataRenamed = self.data.rename(columns=replacementDict)
        if self.datasetID == 'MiD08':
            dataRenamed['hhPersonID'] = (dataRenamed['hhID'].astype('string') +
                                         dataRenamed['personID'].astype('string')).astype('int')
        self.data = dataRenamed
        print('Finished harmonization of variables')

    def createReplacementDict(self, datasetID: str, dictRaw: dict) -> dict:
        """
        Creates the mapping dictionary from raw data variable names to VencoPy internal variable names as specified
        in parseConfig.yaml for the specified data set.

        :param datasetID: list of strings declaring the datasetIDs to be read in
        :param dictRaw: Contains dictionary of the raw data
        :return: Dictionary with internal names as keys and raw data column names as values.
        """

        if datasetID in dictRaw['datasetID']:
            listIndex = dictRaw['datasetID'].index(datasetID)
            return {val[listIndex]: key for (key, val) in dictRaw.items()}
        else:
            raise ValueError(f'Data set {datasetID} not specified in MiD variable dictionary.')

    def convertTypes(self):
        """
        Convert raw column types to predefined python types as specified in parseConfig['inputDTypes']. This is mainly
        done for performance reasons. But also in order to avoid index values that are of type int to be cast to float.
        The function operates only on self.data and writes back changes to self.data

        :return: None
        """

        # Filter for dataset specific columns
        conversionDict = self.parseConfig['inputDTypes']
        keys = {iCol for iCol in conversionDict.keys() if iCol in self.data.columns}
        self.subDict = {key: conversionDict[key] for key in conversionDict.keys() & keys}
        self.data = self.data.astype(self.subDict)

    def returnDictBottomValues(self, baseDict: dict, lst: list = []) -> list:
        """
        Returns a list of all dictionary values of the last dictionary level (the bottom) of baseDict. The parameter
        lst is used as an interface between recursion levels.

        :param baseDict: Dictionary of variables
        :param lst: empty list, is used as interface to next recursion
        :return: Returns a list with all the bottom dictionary values
        """
        for iKey, iVal in baseDict.items():
            if isinstance(iVal, dict):
                lst = self.returnDictBottomValues(iVal, lst)
            else:
                if iVal is not None:
                    lst.append(iVal)
        return lst

    def checkFilterDict(self):
        """
        Checking if all values of filter dictionaries are of type list.  Currently only checking if list of list str
        not typechecked all(map(self.__checkStr, val). Conditionally triggers an assert.

        :return: None
        """

        assert all(isinstance(val, list) for val in self.returnDictBottomValues(self.__filterDict)), \
            f'All values in filter dictionaries have to be lists, but are not'

    def returnDictBottomKeys(self, baseDict: dict, lst: list = None) -> list:
        """
        Returns the lowest level keys of baseDict and returns all of them as a list. The parameter lst is used as
        interface between recursion levels.

        :param baseDict: Dictionary of variables
        :param lst: empty list, used as interface between recursion levels
        :return: Returns a list with all the bottom level dictionary keys
        """

        if lst is None:
            lst = []
        for iKey, iVal in baseDict.items():
            if isinstance(iVal, dict):
                lst = self.returnDictBottomKeys(iVal, lst)
            else:
                if iVal is not None:
                    lst.append(iKey)
        return lst

    def filter(self):
        """
        Wrapper function to carry out filtering for the four filter logics of including, excluding, greaterThan and
        smallerThan. If a filterDict is defined with a different key, a warning is thrown. The function operates on
        self.data class-internally.

        :return: None
        """

        print(f'Starting filtering, applying {len(self.returnDictBottomKeys(self.__filterDict))} filters.')
        ret = pd.DataFrame(index=self.data.index)
        # Future releases: as discussed before we could indeed work here with a plug and pray approach.
        #  we would need to introduce a filter manager and a folder structure where to look for filters.
        #  this is very similar code than the one from ioproc. If we want to go down this route we should
        #  take inspiration from the code there. It was not easy to get it right in the first place. This
        #  might be easy to code but hard to implement correctly.
        for iKey, iVal in self.__filterDict.items():
            if iKey == 'include':
                ret = ret.join(self.setIncludeFilter(iVal, self.data.index))
            elif iKey == 'exclude':
                ret = ret.join(self.setExcludeFilter(iVal, self.data.index))
            elif iKey == 'greaterThan':
                ret = ret.join(self.setGreaterThanFilter(iVal, self.data.index))
            elif iKey == 'smallerThan':
                ret = ret.join(self.setSmallerThanFilter(iVal, self.data.index))
            else:
                warnings.warn(f'A filter dictionary was defined in the parseConfig with an unknown filtering key. '
                              f'Current filtering keys comprise include, exclude, smallerThan and greaterThan.'
                              f'Continuing with ignoring the dictionary {iKey}')
        self.data = self.data[ret.all(axis='columns')]
        self.filterAnalysis(ret)

    def setIncludeFilter(self, includeFilterDict: dict, dataIndex) -> pd.DataFrame:
        """
        Read-in function for include filter dict from parseConfig.yaml

        :param includeFilterDict: Dictionary of include filters defined in parseConfig.yaml
        :param dataIndex: Index for the data frame
        :return: Returns a data frame with individuals using car as a mode of transport
        """

        incFilterCols = pd.DataFrame(index=dataIndex, columns=includeFilterDict.keys())
        for incCol, incElements in includeFilterDict.items():
            incFilterCols[incCol] = self.data[incCol].isin(incElements)
        return incFilterCols

    def setExcludeFilter(self, excludeFilterDict: dict, dataIndex) -> pd.DataFrame:
        """
        Read-in function for exclude filter dict from parseConfig.yaml

        :param excludeFilterDict: Dictionary of exclude filters defined in parseConfig.yaml
        :param dataIndex: Index for the data frame
        :return: Returns a filtered data frame with exclude filters
        """
        exclFilterCols = pd.DataFrame(index=dataIndex, columns=excludeFilterDict.keys())
        for excCol, excElements in excludeFilterDict.items():
            exclFilterCols[excCol] = ~self.data[excCol].isin(excElements)
        return exclFilterCols

    def setGreaterThanFilter(self, greaterThanFilterDict: dict, dataIndex):
        """
        Read-in function for greaterThan filter dict from parseConfig.yaml

        :param greaterThanFilterDict: Dictionary of greater than filters defined in parseConfig.yaml
        :param dataIndex: Index for the data frame
        :return:
        """
        greaterThanFilterCols = pd.DataFrame(index=dataIndex, columns=greaterThanFilterDict.keys())
        for greaterCol, greaterElements in greaterThanFilterDict.items():
            greaterThanFilterCols[greaterCol] = self.data[greaterCol] >= greaterElements.pop()
            if len(greaterElements) > 0:
                warnings.warn(f'You specified more than one value as lower limit for filtering column {greaterCol}.'
                              f'Only considering the last element given in the parseConfig.')
        return greaterThanFilterCols

    def setSmallerThanFilter(self, smallerThanFilterDict: dict, dataIndex) -> pd.DataFrame:
        """
        Read-in function for smallerThan filter dict from parseConfig.yaml

        :param smallerThanFilterDict: Dictionary of smaller than filters defined in parseConfig.yaml
        :param dataIndex: Index for the data frame
        :return: Returns a data frame of trips covering a distance of less than 1000 km
        """
        smallerThanFilterCols = pd.DataFrame(index=dataIndex, columns=smallerThanFilterDict.keys())
        for smallerCol, smallerElements in smallerThanFilterDict.items():
            smallerThanFilterCols[smallerCol] = self.data[smallerCol] <= smallerElements.pop()
            if len(smallerElements) > 0:
                warnings.warn(f'You specified more than one value as upper limit for filtering column {smallerCol}.'
                              f'Only considering the last element given in the parseConfig.')
        return smallerThanFilterCols

    def filterAnalysis(self, filterData: pd.DataFrame):
        """
        Function supplies some aggregate info of the data after filtering to the user Function does not change any
        class attributes

        :param filterData:
        :return: None
        """

        lenData = sum(filterData.all(axis='columns'))
        boolDict = {iCol: sum(filterData[iCol]) for iCol in filterData}
        print(f'The following values were taken into account after filtering:')
        pprint.pprint(boolDict)
        print(f"All filters combined yielded a total of {lenData} was taken into account")
        print(f'This corresponds to {lenData / len(filterData)* 100} percent of the original data')

    def filterConsistentHours(self):
        """
        Filtering out records where starting hour is after end hour but trip takes place on the same day.
        These observations are data errors.

        :return: No returns, operates only on the class instance
        """
        dat = self.data
        self.data = dat.loc[(dat['tripStartClock'] <= dat['tripEndClock']) | (dat['tripEndNextDay'] == 1), :]
        # If we want to get rid of tripStartClock and tripEndClock (they are redundant variables)
        # self.data = dat.loc[pd.to_datetime(dat.loc[:, 'tripStartHour']) <= pd.to_datetime(dat.loc[:, 'tripEndHour']) |
        #                     (dat['tripEndNextDay'] == 1), :]

    def addStrColumns(self, weekday=True, purpose=True):
        """
        Adds string columns for either weekday or purpose.

        :param weekday: Boolean identifier if weekday string info should be added in a separate column
        :param purpose: Boolean identifier if purpose string info should be added in a separate column
        :return: None
        """

        if weekday:
            self.addStrColumnFromMidVariable(colName='weekdayStr', varName='tripStartWeekday')
        if purpose:
            self.addStrColumnFromMidVariable(colName='purposeStr', varName='tripPurpose')

    def addStrColumnFromMidVariable(self, colName: str, varName: str):
        """
        Replaces each occurence of a MID variable e.g. 1,2,...,7 for weekdays with an explicitly mapped string e.g.
        'MON', 'TUE',...,'SUN'.

        :param colName: Name of the column in self.data where the explicit string info is stored
        :param varName: Name of the VencoPy internal variable given in config/parseConfig['dataVariables']
        :return: None
        """
        self.data.loc[:, colName] \
            = self.data.loc[:, varName].replace(self.parseConfig['midReplacements'][varName])

    def composeTimestamp(self, data: pd.DataFrame = None,
                         colYear: str = 'tripStartYear',
                         colWeek: str = 'tripStartWeek',
                         colDay: str = 'tripStartWeekday',
                         colHour: str = 'tripStartHour',
                         colMin: str = 'tripStartMinute',
                         colName: str = 'timestampStart') -> np.datetime64:
        """
        :param data: a data frame
        :param colYear: year of start of a particular trip
        :param colWeek: week of start of a particular trip
        :param colDay: weekday of start of a particular trip
        :param colHour: hour of start of a particular trip
        :param colMin: minute of start of a particular trip
        :param colName:
        :return: Returns a detailed time stamp
        """
        data[colName] = pd.to_datetime(data.loc[:, colYear], format='%Y') + \
                        pd.to_timedelta(data.loc[:, colWeek] * 7, unit='days') + \
                        pd.to_timedelta(data.loc[:, colDay], unit='days') + \
                        pd.to_timedelta(data.loc[:, colHour], unit='hour') + \
                        pd.to_timedelta(data.loc[:, colMin], unit='minute')
        # return data

    def composeStartAndEndTimestamps(self):
        """
        :return: Returns start and end time of a trip
        """
        self.composeTimestamp(data=self.data)  # Starting timestamp
        self.composeTimestamp(data=self.data,  # Ending timestamps
                              colHour='tripEndHour',
                              colMin='tripEndMinute',
                              colName='timestampEnd')
        self.updateEndTimestamp()

    def updateEndTimestamp(self):
        """
        :return:
        """
        endsFollowingDay = self.data['tripEndNextDay'] == 1
        self.data.loc[endsFollowingDay, 'timestampEnd'] = self.data.loc[endsFollowingDay,
                                                                        'timestampEnd'] + pd.offsets.Day(1)


class ParseMID(DataParser):
    # Inherited data class to differentiate between abstract interfaces such as vencopy internal
    # variable namings and data set specific functions such as filters etc. Currently not used (06/14/2021)
    pass



if __name__ == '__main__':
    datasetID = 'MiD17'
    pathLocalPathConfig = Path.cwd().parent / 'config' / 'localPathConfig.yaml'  # pathLib syntax for windows, max, linux compatibility, see https://realpython.com/python-pathlib/ for an intro
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathParseConfig = Path.cwd().parent / 'config' / 'parseConfig.yaml'
    with open(pathParseConfig) as ipf:
        parseConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathGlobalConfig = Path.cwd().parent / 'config' / 'globalConfig.yaml'
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    p = DataParser(localPathConfig=localPathConfig, parseConfig=parseConfig, globalConfig=globalConfig,
                   loadEncrypted=False, datasetID=datasetID)
    print(p.data.head())



__version__ = "0.1.0"
__author__ = 'Niklas Wulff'
__contributors__ = 'Benjamin Fuchs'
__credits__ = 'German Aerospace Center (DLR)'
__license__ = 'BSD-3-Clause'

import os
import pathlib
from setuptools import setup, find_packages


def walkDataFiles(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


dataFilePaths = walkDataFiles('./vencopy/config')
dataFilePaths.extend(walkDataFiles('./vencopy/tutorials'))
long_description = (pathlib.Path(__file__).parent.resolve() / 'README.md').read_text(encoding='utf-8')
setup(
    name='vencopy',
    version='0.1.2',
    description='Vehicle Energy Consumption in Python: A tool to simulate load flexibility of electric vehicle fleets.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/dlr-ve/vencopy',
    author='Niklas Wulff',
    author_email='niklas.wulff@dlr.de',
    license='BSD 3-clause',
    packages=['vencopy', 'vencopy.classes', 'vencopy.scripts'],
    package_data={'': dataFilePaths},
    install_requires=['pandas >= 1.1.1, <= 1.2.5',
                      'ruamel.yaml',
                      'seaborn >= 0.9.0',
                      'sphinx >= 2.3.1',
                      'openpyxl >= 3.0.3',
                      'sphinx_rtd_theme >= 0.5.2',
                      'jupyterlab >= 3.1.0',
                      'Click >= 8.0.1',
                      'pyyaml >= 5.1.2'],
    entry_points={
        'console_scripts': [
            'vencopy = vencopy.__main__:create',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering'],
)

